# -*- coding: utf-8 -*-
import os
import re
import time
import datetime
import random
import traceback
import fs
from ..testcliexception import TestCliException

# 定义种子文件的存放目录
seedFileDir = ""
# 定义种子文件的缓存
seedDataCache = {}


# 返回随机的Boolean类型
def random_boolean(p_arg):
    if p_arg:
        pass  # 这里用来避免p_arg参数无用的编译器警告
    return "1" if (random.randint(0, 1) == 1) else "0"


# 返回一个随机的时间
# 有3个参数：
#   stime   开始时间
#   etime   结束时间
#   frmt    日期格式，可以忽略，默认为"%Y-%m-%d"
def random_date(p_arg):
    if len(p_arg) > 2:
        frmt = str(p_arg[2])
    else:
        frmt = "%Y-%m-%d"
    try:
        stime = datetime.datetime.strptime(str(p_arg[0]), frmt)
        etime = datetime.datetime.strptime(str(p_arg[1]), frmt)
    except ValueError:
        raise TestCliException("Invalid date format [" + str(frmt) + "] for [" + str(p_arg[0]) + "]")
    return (random.random() * (etime - stime) + stime).strftime(frmt)


# 返回一个随机的时间戳
# 有3个参数：
#   stime   开始时间
#   etime   结束时间
#   frmt    日期格式，可以忽略，默认为%H:%M:%S
def random_time(p_arg):
    if len(p_arg) > 2:
        frmt = "%Y-%m-%d " + str(p_arg[2])
    else:
        frmt = "%Y-%m-%d %H:%M:%S"
    try:
        stime = datetime.datetime.strptime("2000-01-01 " + str(p_arg[0]), frmt)
        etime = datetime.datetime.strptime("2000-01-01 " + str(p_arg[1]), frmt)
    except ValueError:
        raise TestCliException("Invalid timestamp format [" + str(frmt) + "] for [" + str(p_arg[0]) + "]")
    return (random.random() * (etime - stime) + stime).strftime(frmt)[11:]


# 返回一个随机的时间戳
# 有3个参数：
#   stime   开始时间
#   etime   结束时间
#   frmt    日期格式，可以忽略，默认为%Y-%m-%d %H:%M:%S
def random_timestamp(p_arg):
    if len(p_arg) > 2:
        frmt = str(p_arg[2])
    else:
        frmt = "%Y-%m-%d %H:%M:%S"
    try:
        stime = datetime.datetime.strptime(str(p_arg[0]), frmt)
        etime = datetime.datetime.strptime(str(p_arg[1]), frmt)
    except ValueError:
        raise TestCliException("Invalid timestamp format [" + str(frmt) + "] for [" + str(p_arg[0]) + "]")
    return (random.random() * (etime - stime) + stime).strftime(frmt)


# 返回系统当前时间,Unix时间戳方式
def current_unixtimestamp(p_arg):
    if p_arg:
        pass
    return str(int(time.mktime(datetime.datetime.now().timetuple())))


# 返回系统当前时间
def current_timestamp(p_arg):
    if len(p_arg) == 1:
        frmt = str(p_arg[0]).strip()
        if len(frmt) == 0:
            frmt = "%Y-%m-%d %H:%M:%S"
    else:
        frmt = "%Y-%m-%d %H:%M:%S"
    return datetime.datetime.now().strftime(frmt)


def random_digits(p_arg):
    n = int(p_arg[0])
    seed = '0123456789'.encode()
    len_lc = len(seed)
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = seed[b % len_lc]
    return ba.decode('ascii')


def random_ascii_uppercase(p_arg):
    n = int(p_arg[0])
    seed = 'ABCDEFGHIJKLMNOPRRSTUVWXYZ'.encode()
    len_lc = len(seed)
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = seed[b % len_lc]
    return ba.decode('ascii')


def random_ascii_lowercase(p_arg):
    n = int(p_arg[0])
    seed = 'abcdefghijklmnopqrstuvwxyz'.encode()
    len_lc = len(seed)
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = seed[b % len_lc]
    return ba.decode('ascii')


def random_ascii_letters(p_arg):
    n = int(p_arg[0])
    seed = 'ABCDEFGHIJKLMNOPRRSTUVWXYZabcdefghijklmnopqrstuvwxyz'.encode()
    len_lc = len(seed)
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = seed[b % len_lc]
    return ba.decode('ascii')


def random_ascii_letters_and_digits(p_arg):
    n = int(p_arg[0])
    seed = 'ABCDEFGHIJKLMNOPRRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'.encode()
    len_lc = len(seed)
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = seed[b % len_lc]
    return ba.decode('ascii')


# 返回一个自增长序列
# 有2个参数：
#   identity_name   自增序列号的名字
#   start           即自正常序列的开始值
#   步长目前没有支持，都是1
def identity(p_arg):
    identity_name = str(p_arg[0])
    start = int(p_arg[1])
    if not hasattr(identity, 'x'):
        identity.x = {}
    if identity_name not in identity.x.keys():
        identity.x[identity_name] = start
    else:
        identity.x[identity_name] = identity.x[identity_name] + 1
    return str(identity.x[identity_name])


# 返回一个自增长的时间戳
# 有4个参数：
#   identity_name   自增序列号的名字
#   stime   开始时间
#     若stime写作current_timestamp, 则自增上从当前时间开始
#   frmt    日期格式， 如%Y-%m-%d %H:%M:%S
#   step    步长，可以用ms,us,s来表示，默认情况下是ms
def identity_timestamp(p_arg):
    identity_name = str(p_arg[0])
    ptime = str(p_arg[1])
    frmt = str(p_arg[2])
    step = str(p_arg[3])
    if not hasattr(identity_timestamp, 'x'):
        identity_timestamp.x = {}
    if identity_name not in identity_timestamp.x.keys():
        if ptime == "current_timestamp":
            identity_timestamp.x[identity_name] = datetime.datetime.now()
        else:
            identity_timestamp.x[identity_name] = datetime.datetime.strptime(ptime, frmt)
    else:
        # 判断步长单位，默认是毫秒，可以是s,ms
        if step.endswith("s"):
            if step.endswith("ms"):
                # end with ms 毫秒
                identity_timestamp.x[identity_name] = (identity_timestamp.x[identity_name] +
                                                       datetime.timedelta(milliseconds=float(step[:-2])))
            elif step.endswith("us"):
                # end with us 微妙
                identity_timestamp.x[identity_name] = (identity_timestamp.x[identity_name] +
                                                       datetime.timedelta(microseconds=float(step[:-2])))
            else:
                # end with s 秒
                identity_timestamp.x[identity_name] = (identity_timestamp.x[identity_name] +
                                                       datetime.timedelta(seconds=float(step[:-1])))
        else:
            identity_timestamp.x[identity_name] = (identity_timestamp.x[identity_name] +
                                                   datetime.timedelta(milliseconds=float(step)))
    return identity_timestamp.x[identity_name].strftime(frmt)


# 缓存seed文件
def load_seed_ache():
    global seedFileDir
    global seedDataCache

    if seedFileDir == "":
        raise TestCliException("To generate random data from seed, Please make sure set seed data dir correct.")
    files = os.listdir(seedFileDir)
    for file in files:
        if not file.endswith(".seed"):
            continue
        seed_name = os.path.basename(file)[:-5]  # 去掉.seed的后缀
        with open(os.path.join(seedFileDir, file), mode='r', encoding="utf-8") as f:
            seed_lines = f.readlines()
        for pos in range(0, len(seed_lines)):
            if seed_lines[pos].endswith("\n"):
                seed_lines[pos] = seed_lines[pos][:-1]  # 去掉回车换行
        seedDataCache[seed_name] = seed_lines


def random_from_seed(p_arg):
    global seedDataCache

    seed_name = str(p_arg[0])
    if len(p_arg) == 3:
        # 从指定位置开始截取数据内容
        start_pos = int(p_arg[1])
        max_length = int(p_arg[2])
    elif len(p_arg) == 2:
        # 从头开始截取文件内容
        start_pos = 0
        max_length = int(p_arg[1])
    else:
        start_pos = 0
        max_length = -1

    # 如果还没有加载种子，就先尝试加载
    if seed_name not in seedDataCache:
        load_seed_ache()

    # 在种子中查找需要的内容
    if seed_name in seedDataCache:
        n = len(seedDataCache[seed_name])
        if n == 0:
            # 空的Seed文件，可能是由于Seed文件刚刚创建，则尝试重新加载数据
            load_seed_ache()
            n = len(seedDataCache[seed_name])
            if n == 0:
                raise TestCliException("Seed cache is zero. [" + str(p_arg[0]) + "].")
        random_lines = seedDataCache[seed_name][random.randint(0, n - 1)]
        if max_length == -1:
            return random_lines[start_pos:]
        else:
            return random_lines[start_pos:start_pos + max_length]
    else:
        raise TestCliException("Unknown seed [" + str(p_arg[0]) + "].  Please create it first.")


# 将传递的SQL字符串转换成一个带有函数指针的数组
def parse_formula_str(p_formula_str):
    # 首先按照{}来把所有的函数定义分拆到数组中
    m_row_struct = re.split('[{}]', p_formula_str)
    m_return_row_struct = []

    # 在分拆的数组中依次查找关键字，作为后续函数替换的需要
    for m_nRowPos in range(0, len(m_row_struct)):
        if re.search('random_ascii_lowercase|random_ascii_uppercase|random_ascii_letters' +
                     '|random_digits|identity|identity_timestamp|random_ascii_letters_and_digits|random_from_seed' +
                     '|random_date|random_timestamp|random_time|random_boolean|'
                     'current_timestamp|current_unixtimestamp|value',
                     m_row_struct[m_nRowPos], re.IGNORECASE):
            m_function_struct = re.split(r'[(,)]', m_row_struct[m_nRowPos])
            for pos in range(0, len(m_function_struct)):
                m_function_struct[pos] = m_function_struct[pos].strip()
                if m_function_struct[pos].startswith("'"):
                    m_function_struct[pos] = m_function_struct[pos][1:]
                if m_function_struct[pos].endswith("'"):
                    m_function_struct[pos] = m_function_struct[pos][0:-1]
            if len(m_function_struct[-1]) == 0:
                m_function_struct.pop()
            m_call_out_struct = []
            if m_function_struct[0].upper() == "RANDOM_ASCII_LOWERCASE":
                m_call_out_struct.append(random_ascii_lowercase)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_ASCII_LOWERCASE", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_ASCII_LOWERCASE", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_ascii_lowercase)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_ASCII_UPPERCASE":
                m_call_out_struct.append(random_ascii_uppercase)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_ASCII_UPPERCASE", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_ASCII_UPPERCASE", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_ascii_uppercase)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_ASCII_LETTERS":
                m_call_out_struct.append(random_ascii_letters)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_ASCII_LETTERS", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_ASCII_LETTERS", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_ascii_letters)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_DIGITS":
                m_call_out_struct.append(random_digits)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_DIGITS", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_DIGITS", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_digits)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "IDENTITY_TIMESTAMP":
                m_call_out_struct.append(identity_timestamp)
                # 如果没有列名，第一个参数信息是identity#加上列号的信息
                if len(m_function_struct) == 5:
                    m_function_struct = m_function_struct[1:]
                elif len(m_function_struct) == 4:
                    m_function_struct[0] = "identity#" + str(m_nRowPos)
                else:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure IDENTITY_TIMESTAMP has correct parameters.")
                m_call_out_struct.append(m_function_struct)
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):IDENTITY_TIMESTAMP", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):IDENTITY_TIMESTAMP", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(identity_timestamp)
                # 如果有列名，第一个参数信息是identity#加上列名的信息
                m_function_struct[0] = "identity#" + column_name
                m_call_out_struct.append(m_function_struct)
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "IDENTITY":
                m_call_out_struct.append(identity)
                # 如果没有列名，第一个参数信息是identity#加上列号的信息
                m_function_struct[0] = "identity#" + str(m_nRowPos)
                m_call_out_struct.append(m_function_struct)
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):IDENTITY", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):IDENTITY", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(identity)
                # 如果有列名，第一个参数信息是identity#加上列名的信息
                m_function_struct[0] = "identity#" + column_name
                m_call_out_struct.append(m_function_struct)
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_ASCII_LETTERS_AND_DIGITS":
                m_call_out_struct.append(random_ascii_letters_and_digits)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_ASCII_LETTERS_AND_DIGITS", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_ASCII_LETTERS_AND_DIGITS", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_ascii_letters_and_digits)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_FROM_SEED":
                m_call_out_struct.append(random_from_seed)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_FROM_SEED", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_FROM_SEED", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_from_seed)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_DATE":
                m_call_out_struct.append(random_date)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_DATE", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_DATE", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_date)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_TIMESTAMP":
                m_call_out_struct.append(random_timestamp)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_TIMESTAMP", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_TIMESTAMP", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_timestamp)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_TIME":
                m_call_out_struct.append(random_time)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_TIME", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_TIME", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_time)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "RANDOM_BOOLEAN":
                m_call_out_struct.append(random_boolean)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):RANDOM_BOOLEAN", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):RANDOM_BOOLEAN", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(random_boolean)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "CURRENT_TIMESTAMP":
                m_call_out_struct.append(current_timestamp)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):CURRENT_TIMESTAMP", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):CURRENT_TIMESTAMP", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(current_timestamp)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "CURRENT_UNIXTIMESTAMP":
                m_call_out_struct.append(current_unixtimestamp)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append("__NO_NAME__")
            elif re.search(r"(.*):CURRENT_UNIXTIMESTAMP", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):CURRENT_UNIXTIMESTAMP", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                m_call_out_struct.append(current_unixtimestamp)
                m_call_out_struct.append(m_function_struct[1:])
                m_call_out_struct.append(column_name)
            elif m_function_struct[0].upper() == "VALUE":
                # 必须用：开头来表示字段名称
                if not m_function_struct[1:][0].startswith(":"):
                    raise TestCliException("Invalid pattern. Please use Value(:ColumnName).")
                else:
                    column_name = m_function_struct[1:][0][1:]
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name.upper():
                            found = True
                            break
                if not found:
                    m_ValidColumn = ""
                    for row in m_return_row_struct:
                        if isinstance(row, list):
                            m_ValidColumn = m_ValidColumn + row[2] + "|"
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is valid.\n" +
                                           "Valid Column=[" + m_ValidColumn + "]")
                m_call_out_struct.append(None)
                m_call_out_struct.append(column_name)
                m_call_out_struct.append("VALUE")
            elif re.search(r"(.*):VALUE", m_function_struct[0].upper()):
                match_obj = re.search(r"(.*):VALUE", m_function_struct[0].upper())
                column_name = match_obj.group(1).upper().strip()
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name:
                            found = True
                            break
                if found:
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is not duplicate.")
                # 必须用：开头来表示字段名称
                if not m_function_struct[1:][0].startswith(":"):
                    raise TestCliException("Invalid pattern. Please use Value(:ColumnName).")
                else:
                    column_name = m_function_struct[1:][0][1:]
                # 检查列名是否已经定义
                found = False
                for row in m_return_row_struct:
                    if isinstance(row, list):
                        if row[2] == column_name.upper():
                            found = True
                            break
                if not found:
                    m_ValidColumn = ""
                    for row in m_return_row_struct:
                        if isinstance(row, list):
                            m_ValidColumn = m_ValidColumn + row[2] + "|"
                    raise TestCliException("Invalid pattern. "
                                           "Please make sure column [" + column_name + "] is valid.\n" +
                                           "Valid Column=[" + m_ValidColumn + "]")
                m_call_out_struct.append(None)
                m_call_out_struct.append(column_name)
                m_call_out_struct.append("VALUE")
            else:
                raise TestCliException("Invalid pattern. parse [" + m_row_struct[m_nRowPos] + "] failed. ")
            m_return_row_struct.append(m_call_out_struct)
        else:
            m_return_row_struct.append(m_row_struct[m_nRowPos])
    return m_return_row_struct


def get_final_string(p_row_struct):
    m_Result = ""
    m_Saved_ColumnData = {}
    for col in p_row_struct:
        if isinstance(col, list):
            if col[2] == "VALUE":
                column_name = col[1]
                m_Value = m_Saved_ColumnData[column_name.upper()]
                m_Result = m_Result + m_Value
            elif col[2] == "__NO_NAME__":
                m_Result = m_Result + col[0](col[1])
            else:
                m_Value = col[0](col[1])  # 根据函数指针计算返回后的实际内容
                m_Saved_ColumnData[col[2]] = m_Value
                m_Result = m_Result + m_Value
        else:
            m_Result = m_Result + col
    return m_Result


class fileHandler:
    def __init__(self):
        self.fp = None

    def open(self, fileName: str, mode: str, encoding: str):
        self.fp = open(file=fileName, mode=mode, encoding=encoding)

    def read(self, nByte: int):
        return self.fp.read(nByte)

    def write(self, content: bytes):
        self.fp.write(content)

    def writeLines(self, buf: list):
        self.fp.writelines(buf)

    def close(self):
        self.fp.close()


class memHandler:
    def __init__(self):
        self.fp = None
        self.fs = None

    def open(self, fileName: str, mode: str, encoding: str):
        from ..globalvar import globalMemFsHandler
        self.fp = globalMemFsHandler.open(path=fileName, mode=mode, encoding=encoding)

    def read(self, nByte: int):
        return self.fp.read(nByte)

    def write(self, content: bytes):
        self.fp.write(content)

    def writeLines(self, buf: list):
        self.fp.writelines(buf)

    def close(self):
        self.fp.close()


fsImplemention = {
    "MEM": memHandler,
    "FS": fileHandler
}


def createFile(p_filetype: str, p_filename, p_formula_str, p_rows, p_encoding='UTF-8'):
    try:
        if p_filetype.upper() not in fsImplemention.keys():
            raise TestCliException("Unknown target file type [" + str(p_filetype) + "]")
        fsHandler = fsImplemention[p_filetype.upper()]()
        fsHandler.open(fileName=p_filename, mode='w', encoding=p_encoding)

        m_row_struct = parse_formula_str(p_formula_str)
        buf = []
        for i in range(0, p_rows):
            buf.append(get_final_string(m_row_struct) + '\n')
            if len(buf) == 100000:  # 为了提高IO效率，每10W条写入文件一次
                fsHandler.writeLines(buf)
                buf = []
        if len(buf) != 0:
            fsHandler.writeLines(buf)
        fsHandler.close()

        # 重置identity的序列号，保证下次从开头开始
        if hasattr(identity, 'x'):
            delattr(identity, 'x')
        if hasattr(identity_timestamp, 'x'):
            delattr(identity_timestamp, 'x')
    except TestCliException as e:
        raise TestCliException(e.message)
    except Exception as e:
        if "TESTCLI_DEBUG" in os.environ:
            print('traceback.print_exc():\n%s' % traceback.print_exc())
            print('traceback.format_exc():\n%s' % traceback.format_exc())
        raise TestCliException(repr(e))


def convertFile(srcFileType: str, srcFileName: str, dstFileType: str, dstFileName: str):
    blockSize = 8192

    try:
        if srcFileType.upper() not in fsImplemention.keys():
            raise TestCliException("Unknown target file type [" + str(srcFileType) + "]")
        srcHandler = fsImplemention[srcFileType.upper()]()
        srcHandler.open(fileName=srcFileName, mode='rb', encoding=None)
        dstHandler = fsImplemention[dstFileType.upper()]()
        dstHandler.open(fileName=dstFileName, mode='wb', encoding=None)
        while True:
            readContents = srcHandler.read(blockSize)
            dstHandler.write(readContents)
            if len(readContents) <= blockSize:
                break
        srcHandler.close()
        dstHandler.close()
    except TestCliException as e:
        raise TestCliException(e.message)
    except Exception as e:
        if "TESTCLI_DEBUG" in os.environ:
            print('traceback.print_exc():\n%s' % traceback.print_exc())
            print('traceback.format_exc():\n%s' % traceback.format_exc())
        raise TestCliException(repr(e))


def executeDataRequest(cls, requestObject):
    global seedFileDir

    if requestObject["action"] == "set":
        if requestObject["option"] == "seedDir":
            # 设置种子文件的目录
            seedFileDir = requestObject["seedDir"]
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": "Data seed dir has been setted to [" + str(seedFileDir) + "]."
            }
            return
    if requestObject["action"] == "create":
        rowCount = requestObject["rowCount"]
        fileType = requestObject["fileType"]
        targetFile = requestObject["targetFile"]
        columnExpression = requestObject["columnExpression"]
        if rowCount > 1:
            # 如果需要的数据不是一行，则描述信息中的回车换行符号均无意义
            columnExpression = columnExpression.replace('\r', '').replace('\n', '')
        try:
            createFile(p_filetype=fileType,
                       p_filename=targetFile,
                       p_formula_str=columnExpression,
                       p_rows=rowCount,
                       p_encoding=cls.testOptions.get("RESULT_ENCODING")
                       )
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": "File [" + str(targetFile) + "] created successful. " + str(rowCount) + " rows generated."
            }
        except TestCliException as te:
            yield {
                "type": "error",
                "message": te.message,
            }

    if requestObject["action"] == "convert":
        try:
            convertFile(
                srcFileType=requestObject["sourceFileType"],
                srcFileName=requestObject["sourceFile"],
                dstFileType=requestObject["targetFileType"],
                dstFileName=requestObject["targetFile"],
            )
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": "File [" + str(requestObject["sourceFile"]) +
                          "] has been converted to [" + requestObject["targetFile"] + "] generated."
            }
        except TestCliException as te:
            yield {
                "type": "error",
                "message": te.message,
            }



# -*- coding: utf-8 -*-
import copy
import os
import re
from collections import namedtuple

# 默认的比较选项
compareDefaultOption = {
    "algorithm": "myers",     # Compare比较算法，有LCS和MYERS
    "output": "console",      # Compare结果输出的位置，默认为输出到控制台
    "mask": False,            # 比较中是否按照正则表达式的方式去比较
    "case": False,            # 比较中是否大小写敏感
    "igblank": False,         # 比较中是否忽略空格
    "trim": False,            # 比较中是否忽略行前行尾的空格
    "workEncoding": "utf-8",  # 工作文件的比对字符集
    "refEncoding": "utf-8"    # 参考文件的比对字符集
}
compareOption = copy.copy(compareDefaultOption)
compareMaskLines = {}        # 比对中进行掩码的内容信息
compareSkipLines = []        # 比对中需要忽略的信息


class DiffException(Exception):
    def __init__(self, message):
        Exception.__init__(self)
        self.message = message


class POSIXCompare:
    CompiledRegexPattern = {}
    ErrorRegexPattern = []

    # These define the structure of the history, and correspond to diff output with
    # lines that start with a space, a + and a - respectively.
    Keep = namedtuple('Keep', ['line', 'lineno'])
    Insert = namedtuple('Insert', ['line', 'lineno'])
    Remove = namedtuple('Remove', ['line', 'lineno'])
    # See frontier in myers_diff
    Frontier = namedtuple('Frontier', ['x', 'history'])

    # 正则表达比较两个字符串
    # p_str1                  原字符串
    # p_str2                  正则表达式
    # p_compare_maskEnabled   是否按照正则表达式来判断是否相等
    # p_compare_ignoreCase    是否忽略匹配中的大小写
    def compare_string(self, p_str1, p_str2,
                       p_compare_maskEnabled=False,
                       p_compare_ignoreCase=False):
        # 如果两个字符串完全相等，直接返回
        if p_str1 == p_str2:
            return True

        # 如果忽略大小写的情况下，两个字符串的大写相同，则直接返回
        if p_compare_ignoreCase:
            if p_str1.upper() == p_str2.upper():
                return True

        # 如果没有启用正则，则直接返回不相等
        if not p_compare_maskEnabled:
            return False

        # 用正则判断表达式是否相等
        try:
            # 对正则表达式进行预先编译
            if p_str2 in self.CompiledRegexPattern:
                m_CompiledPattern = self.CompiledRegexPattern[p_str2]
            else:
                # 如果之前编译过，且没有编译成功，不再尝试编译，直接判断不匹配
                if p_str2 in self.ErrorRegexPattern:
                    return False
                # 如果字符串内容不包含特定的正则字符，则直接失败，不再尝试编译
                # 如何快速判断？
                # 正则编译
                if p_compare_ignoreCase:
                    m_CompiledPattern = re.compile(p_str2, re.IGNORECASE)
                else:
                    m_CompiledPattern = re.compile(p_str2)
                self.CompiledRegexPattern[p_str2] = m_CompiledPattern
            matchObj = re.match(m_CompiledPattern, p_str1)
            if matchObj is None:
                return False
            elif str(matchObj.group()) != p_str1:
                return False
            else:
                return True
        except re.error:
            self.ErrorRegexPattern.append(p_str2)
            # 正则表达式错误，可能是由于这并非是一个正则表达式
            return False

    def compareMyers(self,
                     a_lines,
                     b_lines,
                     a_lineno,
                     b_lineno,
                     p_compare_maskEnabled=False,
                     p_compare_ignoreCase=False):
        # This marks the farthest-right point along each diagonal in the edit
        # graph, along with the history that got it there
        frontier = {1: self.Frontier(0, [])}

        history = []
        a_max = len(a_lines)
        b_max = len(b_lines)
        finished = False
        for d in range(0, a_max + b_max + 1):
            for k in range(-d, d + 1, 2):
                # This determines whether our next search point will be going down
                # in the edit graph, or to the right.
                #
                # The intuition for this is that we should go down if we're on the
                # left edge (k == -d) to make sure that the left edge is fully
                # explored.
                #
                # If we aren't on the top (k != d), then only go down if going down
                # would take us to territory that hasn't sufficiently been explored
                # yet.
                go_down = (k == -d or (k != d and frontier[k - 1].x < frontier[k + 1].x))

                # Figure out the starting point of this iteration. The diagonal
                # offsets come from the geometry of the edit grid - if you're going
                # down, your diagonal is lower, and if you're going right, your
                # diagonal is higher.
                if go_down:
                    old_x, history = frontier[k + 1]
                    x = old_x
                else:
                    old_x, history = frontier[k - 1]
                    x = old_x + 1

                # We want to avoid modifying the old history, since some other step
                # may decide to use it.
                history = history[:]
                y = x - k

                # We start at the invalid point (0, 0) - we should only start building
                # up history when we move off of it.
                if 1 <= y <= b_max and go_down:
                    history.append(self.Insert(b_lines[y - 1], b_lineno[y - 1]))
                elif 1 <= x <= a_max:
                    history.append(self.Remove(a_lines[x - 1], a_lineno[x - 1]))

                # Chew up as many diagonal moves as we can - these correspond to common lines,
                # and they're considered "free" by the algorithm because we want to maximize
                # the number of these in the output.
                while x < a_max and y < b_max and \
                        self.compare_string(a_lines[x], b_lines[y],
                                            p_compare_maskEnabled=p_compare_maskEnabled,
                                            p_compare_ignoreCase=p_compare_ignoreCase):
                    x += 1
                    y += 1
                    history.append(self.Keep(a_lines[x - 1], a_lineno[x - 1]))

                if x >= a_max and y >= b_max:
                    # If we're here, then we've traversed through the bottom-left corner,
                    # and are done.
                    finished = True
                    break
                else:
                    frontier[k] = self.Frontier(x, history)
            if finished:
                break
        compareResult = True
        compareDiffResult = []
        for elem in history:
            if isinstance(elem, self.Keep):
                compareDiffResult.append(" {:>{}} ".format(elem.lineno, 6) + elem.line)
            elif isinstance(elem, self.Insert):
                compareResult = False
                compareDiffResult.append("+{:>{}} ".format(elem.lineno, 6) + elem.line)
            else:
                compareResult = False
                compareDiffResult.append("-{:>{}} ".format(elem.lineno, 6) + elem.line)

        return compareResult, compareDiffResult

    def compareLCS(self,
                   x,
                   y,
                   linenox,
                   linenoy,
                   p_compare_maskEnabled=False,
                   p_compare_ignoreCase=False):
        # LCS问题就是求两个字符串最长公共子串的问题。
        # 解法就是用一个矩阵来记录两个字符串中所有位置的两个字符之间的匹配情况，若是匹配则为1，否则为0。
        # 然后求出对角线最长的1序列，其对应的位置就是最长匹配子串的位置。

        # c           LCS数组
        # x           源数据
        # y           目的数据
        # linenox     源数据行数信息
        # linenoy     目的数据行数信息
        # i           源数据长度
        # j           目的数据长度
        # p_result    比对结果
        next_x = x
        next_y = y
        next_i = len(x) - 1
        next_j = len(y) - 1

        # This is our matrix comprised of list of lists.
        # We allocate extra row and column with zeroes for the base case of empty
        # sequence. Extra row and column is appended to the end and exploit
        # Python's ability of negative indices: x[-1] is the last elem.
        # 构建LCS数组
        c = [[0 for _ in range(len(y) + 1)] for _ in range(len(x) + 1)]
        for i, xi in enumerate(x):
            for j, yj in enumerate(y):
                if self.compare_string(xi, yj,
                                       p_compare_maskEnabled, p_compare_ignoreCase):
                    c[i][j] = 1 + c[i - 1][j - 1]
                else:
                    c[i][j] = max(c[i][j - 1], c[i - 1][j])

        # 开始比较
        compare_result = True
        m_CompareDiffResult = []
        while True:
            if next_i < 0 and next_j < 0:
                break
            elif next_i < 0:
                compare_result = False
                m_CompareDiffResult.append("+{:>{}} ".format(linenoy[next_j], 6) + next_y[next_j])
                next_j = next_j - 1
            elif next_j < 0:
                compare_result = False
                m_CompareDiffResult.append("-{:>{}} ".format(linenox[next_i], 6) + next_x[next_i])
                next_i = next_i - 1
            elif self.compare_string(next_x[next_i], next_y[next_j],
                                     p_compare_maskEnabled, p_compare_ignoreCase):
                m_CompareDiffResult.append(" {:>{}} ".format(linenox[next_i], 6) + next_x[next_i])
                next_i = next_i - 1
                next_j = next_j - 1
            elif c[next_i][next_j - 1] >= c[next_i - 1][next_j]:
                compare_result = False
                m_CompareDiffResult.append("+{:>{}} ".format(linenoy[next_j], 6) + next_y[next_j])
                next_j = next_j - 1
            elif c[next_i][next_j - 1] < c[next_i - 1][next_j]:
                compare_result = False
                m_CompareDiffResult.append("-{:>{}} ".format(linenox[next_i], 6) + next_x[next_i])
                next_i = next_i - 1
        return compare_result, m_CompareDiffResult

    def compare_text_files(self, file1, file2,
                           skipLines=None,
                           maskLines=None,
                           ignoreEmptyLine=False,
                           CompareWithMask=None,
                           CompareIgnoreCase=False,
                           CompareIgnoreTailOrHeadBlank=False,
                           CompareWorkEncoding='UTF-8',
                           CompareRefEncoding='UTF-8',
                           compareAlgorithm='MYERS'):
        if not os.path.isfile(file1):
            raise DiffException('ERROR: File %s does not exist!' % file1)
        if not os.path.isfile(file2):
            raise DiffException('ERROR: File %s does not exist!' % file2)

        # 将比较文件加载到数组
        fileRaw = open(file1, mode='r', encoding=CompareWorkEncoding)
        refFileRaw = open(file2, mode='r', encoding=CompareRefEncoding)
        workFile = open(file1, mode='r', encoding=CompareWorkEncoding)
        refFile = open(file2, mode='r', encoding=CompareRefEncoding)
        fileRawContent = fileRaw.readlines()
        refFileRawContent = refFileRaw.readlines()
        workFileContent = workFile.readlines()
        refFileContent = refFile.readlines()

        # linno用来记录行号，在最后输出打印的时候，显示的是原始文件信息，而不是修正后的信息
        lineno1 = []
        lineno2 = []
        for m_nPos in range(0, len(workFileContent)):
            lineno1.append(m_nPos + 1)
        for m_nPos in range(0, len(refFileContent)):
            lineno2.append(m_nPos + 1)

        # 去掉filecontent中的回车换行
        for m_nPos in range(0, len(workFileContent)):
            if workFileContent[m_nPos].endswith('\n'):
                workFileContent[m_nPos] = workFileContent[m_nPos][:-1]
        for m_nPos in range(0, len(refFileContent)):
            if refFileContent[m_nPos].endswith('\n'):
                refFileContent[m_nPos] = refFileContent[m_nPos][:-1]

        # 去掉fileconent中的首尾空格
        if CompareIgnoreTailOrHeadBlank:
            for m_nPos in range(0, len(workFileContent)):
                workFileContent[m_nPos] = workFileContent[m_nPos].lstrip().rstrip()
            for m_nPos in range(0, len(refFileContent)):
                refFileContent[m_nPos] = refFileContent[m_nPos].lstrip().rstrip()

        # 去除在SkipLine里头的所有内容
        if skipLines is not None:
            m_nPos = 0
            while m_nPos < len(workFileContent):
                bMatch = False
                for pattern in skipLines:
                    if self.compare_string(workFileContent[m_nPos], pattern,
                                           p_compare_maskEnabled=True,
                                           p_compare_ignoreCase=True):
                        workFileContent.pop(m_nPos)
                        lineno1.pop(m_nPos)
                        bMatch = True
                        break
                if not bMatch:
                    m_nPos = m_nPos + 1

            m_nPos = 0
            while m_nPos < len(refFileContent):
                bMatch = False
                for pattern in skipLines:
                    if self.compare_string(refFileContent[m_nPos], pattern,
                                           p_compare_maskEnabled=True,
                                           p_compare_ignoreCase=True):
                        refFileContent.pop(m_nPos)
                        lineno2.pop(m_nPos)
                        bMatch = True
                        break
                if not bMatch:
                    m_nPos = m_nPos + 1

        # 去除所有的空行
        if ignoreEmptyLine:
            m_nPos = 0
            while m_nPos < len(workFileContent):
                if len(workFileContent[m_nPos].strip()) == 0:
                    workFileContent.pop(m_nPos)
                    lineno1.pop(m_nPos)
                else:
                    m_nPos = m_nPos + 1
            m_nPos = 0
            while m_nPos < len(refFileContent):
                if len(refFileContent[m_nPos].strip()) == 0:
                    refFileContent.pop(m_nPos)
                    lineno2.pop(m_nPos)
                else:
                    m_nPos = m_nPos + 1

        # 处理MaskLine中的信息，对ref, Work文件进行替换
        if maskLines is not None:
            m_nPos = 0
            while m_nPos < len(refFileContent):
                for m_SQLMaskPattern, m_SQLMaskTarget in maskLines.items():
                    if re.search(m_SQLMaskPattern, refFileContent[m_nPos], re.IGNORECASE) is not None:
                        refFileContent[m_nPos] = \
                            re.sub(m_SQLMaskPattern, m_SQLMaskTarget, refFileContent[m_nPos], flags=re.IGNORECASE)
                m_nPos = m_nPos + 1
            m_nPos = 0
            while m_nPos < len(workFileContent):
                for m_SQLMaskPattern, m_SQLMaskTarget in maskLines.items():
                    if re.search(m_SQLMaskPattern, workFileContent[m_nPos], re.IGNORECASE) is not None:
                        workFileContent[m_nPos] = \
                            re.sub(m_SQLMaskPattern, m_SQLMaskTarget, workFileContent[m_nPos], flags=re.IGNORECASE)
                m_nPos = m_nPos + 1

        # 输出两个信息
        # 1：  Compare的结果是否存在dif，True/False
        # 2:   Compare的Dif列表. 注意：LCS算法是一个翻转的列表. MYERS算法里头是一个正序列表
        if compareAlgorithm == "MYERS":
            (compareResult, compareResultList) = \
                self.compareMyers(workFileContent, refFileContent,
                                  lineno1, lineno2,
                                  p_compare_maskEnabled=CompareWithMask,
                                  p_compare_ignoreCase=CompareIgnoreCase)
        else:
            (compareResult, compareResultList) = \
                self.compareLCS(workFileContent, refFileContent,
                                lineno1, lineno2,
                                p_compare_maskEnabled=CompareWithMask,
                                p_compare_ignoreCase=CompareIgnoreCase)
            # 翻转数组
            compareResultList = compareResultList[::-1]
        # 随后从数组中补充进入被Skip掉的内容
        workLastPos = 0  # 上次Work文件已经遍历到的位置
        refLastPos = 0  # 上次Ref文件已经遍历到的位置
        newCompareResultList = []
        # 从列表中反向开始遍历， Step=-1
        for row in compareResultList:
            if row.startswith('+'):
                # 当前日志没有，Reference Log中有的
                # 需要注意的是，Ref文件中被跳过的行不会补充进入dif文件
                lineno = int(row[1:7])
                m_AppendLine = "+{:>{}} ".format(lineno, 6) + refFileRawContent[lineno - 1]
                if m_AppendLine.endswith("\n"):
                    m_AppendLine = m_AppendLine[:-1]
                newCompareResultList.append(m_AppendLine)
                refLastPos = lineno
                continue
            elif row.startswith('-'):
                # 当前日志有，但是Reference里头没有的
                lineno = int(row[1:7])
                # 补充填写那些已经被忽略规则略掉的内容，只填写LOG文件中的对应信息
                if lineno > (workLastPos + 1):
                    # 当前日志中存在，但是比较的过程中被Skip掉的内容，要首先补充进来
                    for m_nPos in range(workLastPos + 1, lineno):
                        m_AppendLine = "S{:>{}} ".format(m_nPos, 6) + fileRawContent[m_nPos - 1]
                        if m_AppendLine.endswith("\n"):
                            m_AppendLine = m_AppendLine[:-1]
                        newCompareResultList.append(m_AppendLine)
                m_AppendLine = "-{:>{}} ".format(lineno, 6) + fileRawContent[lineno - 1]
                if m_AppendLine.endswith("\n"):
                    m_AppendLine = m_AppendLine[:-1]
                newCompareResultList.append(m_AppendLine)
                workLastPos = lineno
                continue
            elif row.startswith(' '):
                # 两边都有的
                lineno = int(row[0:7])
                # 补充填写那些已经被忽略规则略掉的内容，只填写LOG文件中的对应信息
                if lineno > (workLastPos + 1):
                    # 当前日志中存在，但是比较的过程中被Skip掉的内容，要首先补充进来
                    for m_nPos in range(workLastPos + 1, lineno):
                        m_AppendLine = "S{:>{}} ".format(m_nPos, 6) + fileRawContent[m_nPos - 1]
                        if m_AppendLine.endswith("\n"):
                            m_AppendLine = m_AppendLine[:-1]
                        newCompareResultList.append(m_AppendLine)
                # 完全一样的内容
                m_AppendLine = " {:>{}} ".format(lineno, 6) + fileRawContent[lineno - 1]
                if m_AppendLine.endswith("\n"):
                    m_AppendLine = m_AppendLine[:-1]
                newCompareResultList.append(m_AppendLine)
                workLastPos = lineno
                refLastPos = refLastPos + 1
                continue
            else:
                # 关闭打开的文件
                if fileRaw:
                    fileRaw.close()
                if refFileRaw:
                    refFileRaw.close()
                if workFile:
                    workFile.close()
                if refFile:
                    refFile.close()
                raise DiffException("Missed line number. Bad compare result. [" + row + "]")

        # 关闭打开的文件
        if fileRaw:
            fileRaw.close()
        if refFileRaw:
            refFileRaw.close()
        if workFile:
            workFile.close()
        if refFile:
            refFile.close()
        return compareResult, newCompareResultList


def executeCompareRequest(requestObject):
    global compareOption
    global compareDefaultOption
    global compareMaskLines
    global compareSkipLines

    action = str(requestObject["action"])
    if action == "set":
        if len(requestObject["compareOptions"]) == 0:
            # 如果Compare Set后面没有跟任何参数，则将当前Compare的配置打印出来
            currentCompareSettings = []
            for key, value in compareOption.items():
                currentCompareSettings.append(
                    [str(key), str(value), str(compareDefaultOption[key])])
            yield {
                "type": "result",
                "title": "Current compare settings:",
                "rows": currentCompareSettings,
                "headers": ["Option", "Current value", "Default value"],
                "columnTypes": None,
                "status": None
            }
            currentSkipSettings = []
            for skipline in compareSkipLines:
                currentSkipSettings.append([skipline])
            if len(currentSkipSettings) != 0:
                yield {
                    "type": "result",
                    "title": "Skiplines:",
                    "rows": currentSkipSettings,
                    "headers": ["SkipPattern"],
                    "columnTypes": None,
                    "status": None
                }
            currentMaskSettings = []
            for key, value in compareMaskLines.items():
                currentMaskSettings.append([str(key), str(value)])
            if len(currentMaskSettings) != 0:
                yield {
                    "type": "result",
                    "title": "Masklines:",
                    "rows": currentMaskSettings,
                    "headers": ["SourcePattern", "TargetPattern"],
                    "columnTypes": None,
                    "status": None
                }
        else:  # compareOptions不为空，即命令中指定了选项
            # 设置Compare的选项
            for key, value in requestObject["compareOptions"].items():
                if key in compareOption:
                    compareOption[key] = value
                else:
                    yield {
                        "type": "error",
                        "message": "unknown compare option [" + str(key) + "]"
                    }
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": None
            }

    if action == "reset":
        compareOption = copy.copy(compareDefaultOption)
        compareMaskLines = {}
        compareSkipLines = []
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": None
        }

    if action == "skip":
        source = str(requestObject["source"])
        if source not in compareSkipLines:
            compareSkipLines.append(source)
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": None
        }

    if action == "noskip":
        source = str(requestObject["source"])
        if source in compareSkipLines:
            compareSkipLines.remove(source)
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": None
        }

    if action == "mask":
        source = str(requestObject["source"])
        target = str(requestObject["target"])
        if source not in compareMaskLines.keys():
            compareMaskLines[source] = target
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": None
        }

    if action == "nomask":
        source = str(requestObject["source"])
        if source in compareMaskLines.keys():
            del compareMaskLines[source]
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": None
        }

    if action == "compare":
        # 设置当前命令的比较选项，不一定和默认或者基础设置的选项相同
        currentCompareOption = copy.copy(compareOption)
        for key, value in requestObject["compareOptions"].items():
            if key in currentCompareOption:
                currentCompareOption[key] = value
        targetFile = str(requestObject["targetFile"])
        referenceFile = str(requestObject["referenceFile"])

        compareHandler = POSIXCompare()
        try:
            compareResult, compareReport = \
                compareHandler.compare_text_files(
                    file1=targetFile,
                    file2=referenceFile,
                    CompareWithMask=currentCompareOption["mask"],
                    CompareIgnoreCase=not currentCompareOption["case"],
                    CompareIgnoreTailOrHeadBlank=not currentCompareOption["trim"],
                    compareAlgorithm=currentCompareOption["algorithm"],
                    ignoreEmptyLine=currentCompareOption["igblank"],
                    skipLines=compareSkipLines,
                    maskLines=compareMaskLines,
                    CompareWorkEncoding=currentCompareOption["workEncoding"],
                    CompareRefEncoding=currentCompareOption["refEncoding"],
                )
            if compareResult:
                yield {
                    "type": "result",
                    "title": None,
                    "rows": None,
                    "headers": None,
                    "columnTypes": None,
                    "status": "Compare successful."
                }
            else:
                compareOutput = []
                if "console" in compareOption["output"]:
                    # 如果要求输出到控制台，则显示比对结果到控制台上
                    for line in compareReport:
                        if line.startswith("-") or line.startswith("+"):
                            tag = line[0]
                            workLineno = ""
                            refLineno = ""
                            if tag == "-":
                                workLineno = line[1:7]
                            if tag == "+":
                                refLineno = line[1:7]
                            compareOutput.append([tag, workLineno, refLineno, line[8:]])
                if "diffFile" in compareOption["output"]:
                    # 如果要求输出到文件，则将比对结果写入文件中
                    baseTargetFile = os.path.splitext(targetFile)[0]
                    diffFile = os.path.join(os.path.dirname(os.path.abspath(targetFile)), baseTargetFile + ".dif")
                    fp = open(file=diffFile, mode="w")
                    for line in compareReport:
                        fp.write(line + "\n")
                    fp.close()
                yield {
                    "type": "result",
                    "title": None,
                    "rows": compareOutput,
                    "headers": ["#", "work line#", "ref line#", "content"],
                    "columnTypes": None,
                    "status": "Compare failed."
                }
        except DiffException as de:
            yield {
                "type": "error",
                "message": str(de.message)
            }



# -*- coding: utf-8 -*-

helpMessage = [
    {
        "topic": 'exit',
        "summary": "exit current script with exitValue (Default is 0) ",
        "synatx": "_EXIT [<exitValue>]",
    },
    {
        "topic": 'quit',
        "summary": "force exit current script with exitValue (Default is 0) ",
        "synatx": "_QUIT [<exitValue>]",
    },
    {
        "topic": 'load',
        "summary": "load external map/driver/plugin files.",
        "synatx":
            '''
            _LOAD JDBCDRIVER 
            {
                NAME = <the unique name of each jdbc driver> |
                CLASS = <name of driver class> |
                FILE = <the drive file location> |
                URL = <jdbc connect url> |
                PROPS = <extra properties for jdbc connect url>
            }
            
            _LOAD PLUGIN <plugin file location> 

            _LOAD MAP <command mapping file location>
        ''',
    },
    {
        "topic": 'ssh',
        "summary": "Remote SSH operation.",
        "synatx":
            '''
            _SSH CONNECT <remote hostname or ip address> WITH USER <username> KEYFILE <ssh key file location>
            _SSH CONNECT <remote hostname or ip address> WITH USER <username> [PASSWORD <user password>]
            _SSH EXECUTE <remote command>
            _SSH DISCONNECT
            _SSH SAVE <user defined session name>
            _SSH RESTORE <user defined session name>
            _SSH SFTP CHMOD <remote file name> <new file permission, 8-base number>
            _SSH SFTP GETCWD
            _SSH SFTP CHDIR <new directory>
            _SSH SFTP CHOWN <target file name> <new user id> <new group id>
            _SSH SFTP MKDIR <target directory name> [<new file permission, 8-base number>]
            _SSH SFTP GET <remote file name> <local file name>
            _SSH SFTP PUT <local file name> <remote file name>
            _SSH SFTP REMOVE <file name to delete>
            _SSH SFTP RENAME <file name to rename> <new remote file name>
            _SSH SFTP LISTDIR <file name to list>
            _SSH SFTP TRUNCATE <file name to trancate> <new file size after truncate>
        ''',
    },
    {
        "topic": 'compare',
        "summary": "Diff test result and reference log.",
        "synatx":
        '''
           _COMPARE <work file name> <reference file name> {MASK | NOMASK | CASE | NOCASE | IGBLANK | NOIGBLANK | TRIM | NOTRIM}
           _COMPARE SKIPLINE <the line expression to skip>
           _COMPARE NOSKIPLINE <the skip line to undefine>
           _COMPARE MASKLINE <<src pattern to mask> >= <target string after mask>>
           _COMPARE NOMASKLINE <the mask line to undefine>
           _COMPARE RESET
           _COMPARE SET {MASK | NOMASK | CASE | NOCASE | IGBLANK | NOIGBLANK | TRIM | NOTRIM}
           _COMPARE SET OUTPUT { CONSOLE | DIFFFILE }
           _COMPARE SET ALGORITHM [LCS | MYERS]
           _COMPARE SET WORK ENCODING <work file codec, default is UTF-8>
           _COMPARE SET REFERENCE ENCODING <reference file codec. default is UTF-8>
        ''',
    },
    {
        "topic": 'echo',
        "summary": "echo some message to file.",
        "synatx":
            '''
                _ECHO <target file name>
                    <echo body. you can input anything>
                _ECHO OFF 
            ''',
    },
    {
        "topic": 'spool',
        "summary": "spool following command and command output to file.",
        "synatx":
            '''
                _SPOOL  <target file name>
                    <command1>
                    <command2>
                    .....
                _SPOOL  OFF 
            ''',
    },
    {
        "topic": 'job',
        "summary": "Run slave script in parallel.",
        "synatx":
            '''
               _JOB MANAGER [ON | OOF]
               _JOB WAIT [<job name> | ALL] 
               _JOB SHOW [<job name> | ALL] 
               _JOB START [<job name> | ALL]  
               _JOB ABORT [<job name> | ALL]
               _JOB SHUTDOWN [<job name> | ALL]   
               _JOB CREATE <job name> { <option Name> = <optionValue>}
               _JOB SET <job name> { <option Name> = <optionValue>}
               _JOB REGISTER WORKER TO <job name>
               _JOB DEREGISTER WORKER
               
               Job Option:
               script                   ：  Required. The script name of the background job.    
               tag                      ：  Required. All worker processes with the same tag will be synchronized when 
                                            judging the aggregation point.
               parallel                 ：  Optional. Number of background jobs at same time, default is 1. 
               loop                     ：  Optional. Number of cycles, default is 1.
               timeout                  ：  Optional. The timeout limit (second) of background jobs, 
                                            default value is 0, which means no limit.
               starter_maxprocess       ：  Optional. To reduce the load pressure of the first start, we don't start all 
                                            jobs in a single batch. This parameter is meaningful only when the job first 
                                            started. The default is 9999, which means no limit.
                                            For example: parallel is 10, and starter_maxprocess is 2.
                                                         two jobs are started each "starter_interval" time until the 
                                                         parallel requirements(10) are met.
               starter_interval         ：  Optional. Referring to the previous description, starter_interval define the 
                                            interval time when job first start. default is 0, it means no wait.
               think_time               ：  Optional. After each job completed, the time interval required to start the 
                                            next job. Default is 0, which means no think, start next job immediate.
               blowout_threshold_count  ：  Optional. The failure threshold value. If threshold has reached, we think 
                                            following jobs are unnecessary to run.  The default is 0, that is, no limit.  
            ''',
    },
    {
        "topic": 'data',
        "summary": "Generate test random data.",
        "synatx":
            '''
                _DATA SET SEEDFILE DIR <seed file location>;
                  
                _DATA CREATE MEM|FS FILE <target file location> 
                (
                   <column expression..>,
                   <column expression..>
                )
                [ROWS <rows will generated>];
                  
                _DATA CONVERT MEM|FS FILE <source file location> TO MEM|FS FILE <target file location>;  
            ''',
    },
    {
        "topic": 'SLEEP',
        "summary": "sleep app some time",
        "synatx":
            '''
                _SLEEP <sleep time (seconds)>
            ''',
    },
    {
        "topic": 'ASSERT',
        "summary": "Execute the assertion. Determine whether the specified conditions are met.",
        "synatx":
            '''
                _ASSERT {% <assert python expression> %}, [<name of assertion, optional>]
            ''',
    },
    {
        "topic": 'USE',
        "summary": "Switch the namespace of the current script.",
        "synatx":
            '''
                _USE SQL|API
            ''',
    },
    {
        "topic": 'HOST',
        "summary": "Execute local system commands.",
        "synatx":
            '''
                _HOST <os file command>
            ''',
    },
    {
        "topic": 'SET',
        "summary": "Set/View app runtime options.",
        "synatx":
            '''
                _SET [<OPTION_NAME> <OPTION_VALUE>]
                
                If you run this command without any parameter, it will show current runtime options.
            ''',
    },
    {
        "topic": 'START',
        "summary": "Run sub command script.",
        "synatx":
            '''
                _STARRT <command script file location> [<argv1> <argv2> ....]
            ''',
    },
    {
        "topic": 'SCRIPT',
        "summary": "Run embedded python script.",
        "synatx":
            '''
                > {%
                <python script>
                %}
            ''',
    },
    {
        "topic": 'SPOOL',
        "summary": "Print subsequent run commands and results to the specified file.",
        "synatx":
            '''
                _SPOOL <spool file location>
                <command1 ...>
                <command2 ...>
                _SPOOL OFF
            ''',
    },
]



# -*- coding: utf-8 -*-
import importlib.util
import os
import re
from ..testcliexception import TestCliException
from ..commands.embeddScript import localEmbeddScriptScope

pluginModule = {}
pluginFunction = {}


def loadPlugin(cls, pluginFile: str):
    global pluginFunction
    global pluginModule

    # 首先在当前目录下查找
    # 如果找不到，则去程序所在的目录下去寻找
    if not os.path.exists(pluginFile):
        if not os.path.exists(os.path.join(os.path.abspath(__file__), pluginFile)):
            yield {
                "type": "error",
                "message": "plugin file [" + os.path.abspath(pluginFile) + "] does not exist!",
            }
            return
    pluginFileHandler = open(file=pluginFile, mode="r", encoding=cls.testOptions.get("SCRIPT_ENCODING"))
    pluginContents = pluginFileHandler.readlines()

    # 导入plugin的外部模块
    for pluginContent in pluginContents:
        matchObj = re.match(r"^class(\s+)(.*?)(\s+)?[(|:]", pluginContent)
        if matchObj:
            className = str(matchObj.group(2)).strip()
            # 动态导入class
            spec_class = importlib.util.spec_from_file_location(
                name=className,
                location=os.path.abspath(pluginFile)
            )
            module_class = importlib.util.module_from_spec(spec_class)
            spec_class.loader.exec_module(module_class)
            pluginClass = getattr(module_class, className)
            pluginModule[className] = pluginClass
            localEmbeddScriptScope[className] = pluginModule[className]
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": 'Plugin module [' + str(className) + '] loaded successful.'
            }

    # 导入plugin的函数
    for pluginContent in pluginContents:
        matchObj = re.match(r"^def(\s+)(.*?)(\s+)?[(|:]", pluginContent)
        if matchObj:
            funcName = str(matchObj.group(2)).strip()
            # 动态导入Function
            spec_function = importlib.util.spec_from_file_location(
                name=funcName,
                location=os.path.abspath(pluginFile)
            )
            module_function = importlib.util.module_from_spec(spec_function)
            spec_function.loader.exec_module(module_function)
            pluginFunc = getattr(module_function, funcName)
            pluginFunction[funcName] = pluginFunc
            localEmbeddScriptScope[funcName] = pluginFunction[funcName]
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": 'Plugin function [' + str(funcName) + '] loaded successful.'
            }
    yield {
        "type": "result",
        "title": None,
        "rows": None,
        "headers": None,
        "columnTypes": None,
        "status": 'Plugin file loaded successful.'
    }


# 加载数据库驱动
# 标准的默认驱动程序并不需要使用这个函数，这个函数是用来覆盖标准默认驱动程序的加载信息
def loadJDBCDriver(cls,
                   driverName: str, driverClass: str,
                   driverFile: str, driverURL: str,
                   driverProps: str):
    if driverName is None:  # 显示当前的Driver配置
        m_Result = []
        for row in cls.db_connectionConf:
            m_Result.append([row["Database"], row["ClassName"], row["FullName"],
                             row["JDBCURL"], row["JDBCProp"]])
        yield {
            "type": "result",
            "title": "Current Drivers: ",
            "rows": m_Result,
            "headers": ["Database", "ClassName", "FileName", "JDBCURL", "JDBCProp"],
            "columnTypes": None,
            "status": "Driver loaded."
        }
        return

    # 如果给定的一个全路径，则使用全路径文件, 否则按照相对路径查找
    driverFileFullPathList = []
    if driverFile is not None:
        driverFileList = driverFile.split(',')
        for driverFile in driverFileList:
            driverFile = driverFile.strip().strip('"').strip("'")
            if not os.path.exists(driverFile):
                raise TestCliException("Driver not loaded. "
                                       "file [" + str(os.path.abspath(driverFile)) + "] does not exist!")
            else:
                driverFileFullPathList.append(os.path.abspath(driverFile))

    # 查看指定的驱动是否存在，如果存在，则替换，不存在，则新增
    found = False
    for nPos in range(0, len(cls.db_connectionConf)):
        if cls.db_connectionConf[nPos]["Database"].upper() == driverName.strip().upper():
            m_Config = cls.db_connectionConf[nPos]
            if len(driverFileFullPathList) != 0:
                m_Config["FullName"] = driverFileFullPathList
            if driverClass is not None:
                m_Config["ClassName"] = driverClass
            if driverURL is not None:
                m_Config["JDBCURL"] = driverURL
            if driverProps is not None:
                m_Config["JDBCProp"] = driverProps
            found = True
            cls.db_connectionConf[nPos] = m_Config
    if not found:
        cls.db_connectionConf.append(
            {
                "ClassName": driverClass,
                "FullName": driverFileFullPathList,
                "JDBCURL": driverURL,
                "JDBCProp": driverProps,
                "Database": driverName
            }
        )
    yield {
        "type": "result",
        "title": None,
        "rows": None,
        "headers": None,
        "columnTypes": None,
        "status": "Driver [" + driverName + "] loaded."
    }


# 加载命令行映射信息
def loadMap(cls, mapFile: str):
    try:
        cls.cmdMappingHandler.loadCommandMappings(
            commandScriptFileName=cls.executeScript,
            commandMappings=mapFile
        )
        cls.commandMap = mapFile
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": 'Mapping file loaded.'
        }
    except TestCliException as te:
        yield {
            "type": "error",
            "message": te.message
        }



# Generated from APIParser.g4 by ANTLR 4.11.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .APIParser import APIParser
else:
    from APIParser import APIParser

# This class defines a complete generic visitor for a parse tree produced by APIParser.

class APIParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by APIParser#prog.
    def visitProg(self, ctx:APIParser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#command.
    def visitCommand(self, ctx:APIParser.CommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#http.
    def visitHttp(self, ctx:APIParser.HttpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMessage.
    def visitHttpMessage(self, ctx:APIParser.HttpMessageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpRequestLine.
    def visitHttpRequestLine(self, ctx:APIParser.HttpRequestLineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMethod.
    def visitHttpMethod(self, ctx:APIParser.HttpMethodContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpRequestTarget.
    def visitHttpRequestTarget(self, ctx:APIParser.HttpRequestTargetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpHeaderFields.
    def visitHttpHeaderFields(self, ctx:APIParser.HttpHeaderFieldsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpHeaderField.
    def visitHttpHeaderField(self, ctx:APIParser.HttpHeaderFieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpHeaderFieldName.
    def visitHttpHeaderFieldName(self, ctx:APIParser.HttpHeaderFieldNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpHeaderFieldValue.
    def visitHttpHeaderFieldValue(self, ctx:APIParser.HttpHeaderFieldValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMessageBody.
    def visitHttpMessageBody(self, ctx:APIParser.HttpMessageBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMultipart.
    def visitHttpMultipart(self, ctx:APIParser.HttpMultipartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMultipartBoundary.
    def visitHttpMultipartBoundary(self, ctx:APIParser.HttpMultipartBoundaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpBoundaryDelimiter.
    def visitHttpBoundaryDelimiter(self, ctx:APIParser.HttpBoundaryDelimiterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMessageBodyContent.
    def visitHttpMessageBodyContent(self, ctx:APIParser.HttpMessageBodyContentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMessageBodyOperate.
    def visitHttpMessageBodyOperate(self, ctx:APIParser.HttpMessageBodyOperateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#httpMessageBodyOther.
    def visitHttpMessageBodyOther(self, ctx:APIParser.HttpMessageBodyOtherContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#baseCommand.
    def visitBaseCommand(self, ctx:APIParser.BaseCommandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#exit.
    def visitExit(self, ctx:APIParser.ExitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#quit.
    def visitQuit(self, ctx:APIParser.QuitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#use.
    def visitUse(self, ctx:APIParser.UseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#sleep.
    def visitSleep(self, ctx:APIParser.SleepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#start.
    def visitStart(self, ctx:APIParser.StartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#load.
    def visitLoad(self, ctx:APIParser.LoadContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#assert.
    def visitAssert(self, ctx:APIParser.AssertContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#host.
    def visitHost(self, ctx:APIParser.HostContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#loop.
    def visitLoop(self, ctx:APIParser.LoopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#if.
    def visitIf(self, ctx:APIParser.IfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#endif.
    def visitEndif(self, ctx:APIParser.EndifContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#whenever.
    def visitWhenever(self, ctx:APIParser.WheneverContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#spool.
    def visitSpool(self, ctx:APIParser.SpoolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#echo.
    def visitEcho(self, ctx:APIParser.EchoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#script.
    def visitScript(self, ctx:APIParser.ScriptContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#set.
    def visitSet(self, ctx:APIParser.SetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#ssh.
    def visitSsh(self, ctx:APIParser.SshContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#job.
    def visitJob(self, ctx:APIParser.JobContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#compare.
    def visitCompare(self, ctx:APIParser.CompareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#data.
    def visitData(self, ctx:APIParser.DataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#help.
    def visitHelp(self, ctx:APIParser.HelpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by APIParser#monitor.
    def visitMonitor(self, ctx:APIParser.MonitorContext):
        return self.visitChildren(ctx)



del APIParser


# -*- coding: utf-8 -*-
import os
import re
import codecs
import time
import traceback
from ..globalvar import localEmbeddScriptScope


# 从文件中执行SQL
def executeFile(cls, scriptFile, argv):
    # 将scriptFile根据平台进行转义
    try:
        if str(scriptFile).startswith("'"):
            scriptFile = scriptFile[1:]
        if str(scriptFile).endswith("'"):
            scriptFile = scriptFile[:-1]
        with open(os.path.expanduser(scriptFile), encoding=cls.testOptions.get("SCRIPT_ENCODING")) as f:
            query = f.read()

        # 空文件直接返回
        if len(query) == 0:
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": None
            }
            return

        # 将程序的参数记录到环境信息中
        localEmbeddScriptScope["argv"] = argv

        # 处理脚本文件头数据, 文件开头的0xFEFF,codecs.BOM_UTF8忽略不看
        if ord(query[0]) == 0xFEFF:
            # 去掉脚本文件可能包含的UTF-BOM
            query = query[1:]
        if query[:3] == codecs.BOM_UTF8:
            # 去掉脚本文件可能包含的UTF-BOM
            query = query[3:]

        # Scenario, Transaction等Hint信息不会带入到下一个脚本文件中
        cls.cmdExecuteHandler.scenario = ''
        cls.cmdExecuteHandler.transaction = ''
        cls.cmdExecuteHandler.setStartTime(time.time())

        # 这里需要把command按照namespace分离，即不同的namespace分开执行
        # 避免切换namespace后，解析导致的问题
        # 其中：
        #   __use__ namespace 为一段
        #   两个 __use__ namespace之间为一段
        commandList = []
        lastCommand = None
        lastNameSpace = cls.nameSpace
        for commandLine in query.split("\n"):
            match_obj = re.match(r"(\s+)?__USE__(\s+)NAMESPACE(\s+)(.*)$",
                                 commandLine, re.IGNORECASE | re.DOTALL)
            if match_obj:
                newNameSpace = match_obj.group(4).strip().upper()
                if lastCommand is not None:
                    commandList.append({"nameSpace": lastNameSpace, "script": lastCommand})
                    lastCommand = None
                commandList.append({"nameSpace": "GENERAL", "script": commandLine})
                lastNameSpace = newNameSpace
                lastNameSpace = lastNameSpace.strip().upper().strip(';')
                continue
            else:
                if lastCommand is None:
                    lastCommand = commandLine
                else:
                    lastCommand = lastCommand + "\n" + commandLine
        if lastCommand is not None:
            commandList.append({"nameSpace": lastNameSpace, "script": lastCommand})

        # 分段执行执行的语句
        for command in commandList:
            for executeResult in \
                    cls.cmdExecuteHandler.runStatement(
                        statement=command["script"],
                        commandScriptFile=os.path.expanduser(scriptFile),
                        nameSpace=command["nameSpace"]
                    ):
                yield executeResult
    except IOError as e:
        if "TESTCLI_DEBUG" in os.environ:
            print('traceback.print_exc():\n%s' % traceback.print_exc())
            print('traceback.format_exc():\n%s' % traceback.format_exc())
        yield {
            "type": "result",
            "title": None,
            "rows": None,
            "headers": None,
            "columnTypes": None,
            "status": "Execute script [" + str(os.path.abspath(scriptFile)) + "] failed. " + repr(e)
        }



# -*- coding: utf-8 -*-
import datetime
import time
from json import JSONDecodeError

import click
import json
import os
import re
import platform
import binascii
import decimal
import traceback
import urllib3
from .sqlclijdbc import SQLCliJDBCException
from .sqlclijdbc import SQLCliJDBCLargeObject
from .sqlclijdbc import SQLCliJDBCTimeOutException
from .sqlparse import SQLAnalyze
from .apiparse import APIAnalyze
from .sqlparse import SQLFormatWithPrefix
from .apiparse import APIRequestStringFormatWithPrefix

from .commands.load import loadPlugin
from .commands.load import loadJDBCDriver
from .commands.load import loadMap
from .commands.exit import exitApplication
from .commands.session import sessionManage
from .commands.assertExpression import assertExpression
from .commands.assertExpression import evalExpression
from .commands.embeddScript import executeEmbeddScript
from .commands.connectdb import connectDb, disconnectDb
from .commands.start import executeFile
from .commands.host import executeLocalCommand
from .commands.spool import spool
from .commands.echo import echo_input
from .commands.setOptions import setOptions
from .commands.cliSleep import cliSleep
from .commands.userNameSpace import userNameSpace
from .commands.whenever import setWheneverAction
from .commands.ssh import executeSshRequest
from .commands.ssh import rewriteSshRequest
from .commands.compare import executeCompareRequest
from .commands.helper import showHelp
from .commands.data import executeDataRequest
from .commands.monitor import executeMonitorRequest

from .common import rewriteHintStatement
from .common import rewriteSQLStatement
from .common import rewriteAPIStatement
from .common import parseSQLHints
from .common import parseAPIHints
from .common import sortresult
from .testcliexception import TestCliException
from .globalvar import lastCommandResult


class CmdExecute(object):
    def __init__(self):
        # 程序处理选项
        self.testOptions = None

        # 主程序处理句柄
        self.cliHandler = None

        # 当前Executeor的WorkerName
        self.workerName = None

        # 程序Echo输出句柄
        self.echofile = None

        # 脚本启动的时间
        self.startTime = time.time()

        # 当前执行的命令脚本
        self.script = None

        # 命令重写处理
        self.mappingHandler = None

        # Scenario名称，如果当前未指定，则重复上一个命令的Scenario信息
        self.scenarioId = ''
        self.scenarioName = ''

        # 当前脚本的TimeOut设置
        self.sqlTimeOut = -1          # SQL执行的超时时间设置
        self.apiTimeOut = -1          # API执行的超时时间设置
        self.scriptTimeOut = -1       # 脚本执行的超时时间设置
        self.timeout = -1             # 当前SQL的超时时间设置
        self.timeOutMode = None       # COMMAND|SCRIPT|NONE

        # 数据库连接
        self.sqlConn = None
        self.sqlCursor = None
        self.apiConn = None

        # 记录语句是否当前处于条件表达式判定中
        self.ifMode = False
        self.ifCondition = False
        # 记录语句是否在循环过程中
        self.loopMode = False
        self.loopCondition = False
        self.loopStartPos = 0

    def setStartTime(self, startTime):
        self.startTime = startTime

    def getStartTime(self):
        return self.startTime

    def jqparse(self, obj, path='.'):
        class DecimalEncoder(json.JSONEncoder):
            def default(self, o):
                if isinstance(o, decimal.Decimal):
                    return float(o)
                super(DecimalEncoder, self).default(o)

        if self is None:
            pass
        if obj is None:
            if "TESTCLI_DEBUG" in os.environ:
                click.secho("[DEBUG] JQ Parse Error: obj is None")
            return "****"
        try:
            obj = json.loads(obj) if isinstance(obj, str) else obj
            find_str, find_map = '', ['["%s"]', '[%s]', '%s', '.%s']
            for im in path.split('.'):
                if not im:
                    continue
                if obj is None:
                    if "TESTCLI_DEBUG" in os.environ:
                        click.secho("[DEBUG] JQ Parse Error: obj is none")
                    return "****"
                if isinstance(obj, (list, tuple, str)):
                    if im.startswith('[') and im.endswith(']'):
                        im = im[1:-1]
                    if ':' in im:
                        slice_default = [0, len(obj), 1]
                        obj, quota = obj[slice(
                            *[int(sli) if sli else slice_default[i] for i, sli in
                              enumerate(im.split(':'))])], 1
                    else:
                        obj, quota = obj[int(im)], 1
                else:
                    if im in obj:
                        obj, quota = obj.get(im), 0
                    elif im.endswith('()'):
                        obj, quota = list(getattr(obj, im[:-2])()), 3
                    else:
                        if im.isdigit():
                            obj, quota = obj[int(im)], 1
                        else:
                            raise KeyError(im)
                find_str += find_map[quota] % im
            return obj if isinstance(obj, str) else json.dumps(obj,
                                                               sort_keys=True,
                                                               ensure_ascii=False,
                                                               cls=DecimalEncoder)
        except (IndexError, KeyError, ValueError) as je:
            if "TESTCLI_DEBUG" in os.environ:
                click.secho("[DEBUG] JQ Parse Error: " + repr(je))
            return "****"

    def getcommandResult(self, cursor, rowcount):
        """
            返回的内容：
                title           输出的前提示信息
                result          结果数据集
                headers         表头信息
                columnTypes     结果字段类型
                status          输出的后提示信息
                FetchStatus     是否输出完成
                rowcount        共返回记录行数
                Warning         警告信息
        """
        title = headers = None
        fetchStatus = True

        def format_column(p_column, p_columntype):
            if type(p_column) == float:
                return self.testOptions.get("FLOAT_FORMAT") % p_column
            elif type(p_column) in (bool, str, int):
                return p_column
            elif type(p_column) == list:
                return p_column
            elif type(p_column) == datetime.date:
                columnFormat = self.testOptions.get("DATE_FORMAT")
                if platform.system().lower() in ['windows', 'darwin']:
                    columnFormat = columnFormat.replace("%04Y", "%Y")
                else:
                    columnFormat = columnFormat.replace("%Y", "%04Y")
                return p_column.strftime(columnFormat)
            elif type(p_column) == datetime.datetime:
                if p_columntype in ["TIMESTAMP WITH TIME ZONE",
                                    "TIMESTAMP WITH LOCAL TIME ZONE"]:
                    columnFormat = self.testOptions.get("DATETIME-TZ_FORMAT")
                else:
                    columnFormat = self.testOptions.get("DATETIME_FORMAT")
                if platform.system().lower() in ['windows', 'darwin']:
                    columnFormat = columnFormat.replace("%04Y", "%Y")
                else:
                    columnFormat = columnFormat.replace("%Y", "%04Y")
                return p_column.strftime(columnFormat)
            elif type(p_column) == datetime.time:
                return p_column.strftime(self.testOptions.get("TIME_FORMAT"))
            elif type(p_column) == bytearray:
                if p_columntype == "BLOB":
                    columnTrimedLength = int(self.testOptions.get("LOB_LENGTH"))
                    bColumnFullOutput = True
                    if len(p_column) > columnTrimedLength:
                        bColumnFullOutput = False
                        p_column = p_column[:columnTrimedLength]
                    # 转换为16进制，并反算成ASCII
                    p_column = binascii.b2a_hex(p_column)
                    p_column = p_column.decode()
                    if not bColumnFullOutput:
                        # 用...的方式提醒输出没有结束，只是由于格式控制导致不显示
                        return "0x" + p_column + "..."
                    else:
                        return "0x" + p_column
                else:
                    # 转换为16进制，并反算成ASCII
                    p_column = binascii.b2a_hex(p_column)
                    p_column = p_column.decode()
                    return "0x" + p_column
            elif type(p_column) == decimal.Decimal:
                if self.testOptions.get("DECIMAL_FORMAT") != "":
                    return self.testOptions.get("DECIMAL_FORMAT") % p_column
                else:
                    return p_column
            elif type(p_column) == SQLCliJDBCLargeObject:
                trimedLength = int(self.testOptions.get("LOB_LENGTH"))
                if trimedLength < 4:
                    trimedLength = 4
                if trimedLength > p_column.getObjectLength():
                    if p_column.getColumnTypeName().upper().find("CLOB") != -1:
                        dataValue = p_column.getData(1, p_column.getObjectLength())
                        return dataValue
                    elif p_column.getColumnTypeName().upper().find("BLOB") != -1:
                        dataValue = p_column.getData(1, p_column.getObjectLength())
                        dataValue = binascii.b2a_hex(dataValue)
                        dataValue = dataValue.decode()
                        return "0x" + dataValue
                else:
                    if p_column.getColumnTypeName().upper().find("CLOB") != -1:
                        dataValue = "Len:" + str(p_column.getObjectLength()) + ";" + \
                                      "Content:[" + \
                                      p_column.getData(1, trimedLength - 3) + "..." + \
                                      p_column.getData(p_column.getObjectLength() - 2, 3) + \
                                      "]"
                        return dataValue
                    elif p_column.getColumnTypeName().upper().find("BLOB") != -1:
                        dataValue = "Len:" + str(p_column.getObjectLength()) + ";" + \
                                      "Content:0x[" + \
                                      binascii.b2a_hex(p_column.getData(1, trimedLength - 3)).decode() + "..." + \
                                      binascii.b2a_hex(p_column.getData(p_column.getObjectLength() - 2, 3)).decode() + \
                                      "]"
                        return dataValue
            elif isinstance(p_column, type(None)):
                return p_column
            else:
                # 其他类型直接返回
                raise SQLCliJDBCException("TestCli-0000: Unknown column type [" +
                                          str(p_columntype) + ":" + str(type(p_column)) +
                                          "] in format_column")

        # cursor.description is not None for queries that return result sets,
        # e.g. SELECT.
        result = []
        columnTypes = []
        if cursor.description is not None:
            headers = [x[0] for x in cursor.description]
            columnTypes = [x[1] for x in cursor.description]
            if cursor.warnings is not None:
                status = "{0} row{1} selected with warnings."
            else:
                status = "{0} row{1} selected."
            arraySize = int(self.testOptions.get("SQL_FETCHSIZE"))
            rowset = cursor.fetchmany(arraySize)
            for row in rowset:
                collatedRow = []
                for nColumnPos in range(0, len(row)):
                    column = row[nColumnPos]
                    columntype = columnTypes[nColumnPos]
                    # 对于空值直接返回
                    if column is None:
                        collatedRow.append(None)
                        continue

                    # 处理各种数据类型
                    if columnTypes[nColumnPos] == "STRUCT":
                        columnValue = "STRUCTURE("
                        for pos in range(0, len(column)):
                            m_ColumnType = str(type(column[pos]))
                            if pos == 0:
                                if type(column[pos]) == str:
                                    columnValue = columnValue + "'" + str(column[pos]) + "'"
                                elif type(column[pos]) == datetime.date:
                                    columnValue = columnValue + "DATE '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif type(column[pos]) == datetime.datetime:
                                    columnValue = columnValue + "TIMESTAMP '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif isinstance(column[pos], type(None)):
                                    columnValue = columnValue + "<null>"
                                else:
                                    columnValue = columnValue + \
                                                    str(format_column(column[pos], m_ColumnType))
                            else:
                                if type(column[pos]) == str:
                                    columnValue = columnValue + ",'" + str(column[pos]) + "'"
                                elif type(column[pos]) == datetime.date:
                                    columnValue = columnValue + ",DATE '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif type(column[pos]) == datetime.datetime:
                                    columnValue = columnValue + ",TIMESTAMP '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif isinstance(column[pos], type(None)):
                                    columnValue = columnValue + ",<null>"
                                else:
                                    columnValue = columnValue + "," + \
                                                    str(format_column(column[pos], m_ColumnType))
                        columnValue = columnValue + ")"
                        collatedRow.append(columnValue)
                    elif columnTypes[nColumnPos] == "ARRAY":
                        columnValue = "ARRAY["
                        if self.testOptions.get('OUTPUT_SORT_ARRAY') == "ON":
                            # 保证Array的输出每次都一样顺序
                            # 需要注意可能有NULL值导致字符数组无法排序的情况, column是一个一维数组
                            column.sort(key=lambda x: (x is None, x))
                        for pos in range(0, len(column)):
                            m_ColumnType = str(type(column[pos]))
                            if pos == 0:
                                if type(column[pos]) == str:
                                    columnValue = columnValue + "'" + str(column[pos]) + "'"
                                elif type(column[pos]) == datetime.date:
                                    columnValue = columnValue + "DATE '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif type(column[pos]) == datetime.datetime:
                                    columnValue = columnValue + "TIMESTAMP '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif isinstance(column[pos], type(None)):
                                    columnValue = columnValue + "<null>"
                                else:
                                    columnValue = columnValue + \
                                                    str(format_column(column[pos], m_ColumnType))
                            else:
                                if type(column[pos]) == str:
                                    columnValue = columnValue + ",'" + str(column[pos]) + "'"
                                elif type(column[pos]) == datetime.date:
                                    columnValue = columnValue + ",DATE '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif type(column[pos]) == datetime.datetime:
                                    columnValue = columnValue + ",TIMESTAMP '" + \
                                                    format_column(column[pos], m_ColumnType) + "'"
                                elif isinstance(column[pos], type(None)):
                                    columnValue = columnValue + ",<null>"
                                else:
                                    columnValue = columnValue + "," + \
                                                    str(format_column(column[pos], m_ColumnType))
                        columnValue = columnValue + "]"
                        collatedRow.append(columnValue)
                    else:
                        collatedRow.append(format_column(column, columntype))
                collatedRow = tuple(collatedRow)
                result.append(collatedRow)
            rowcount = rowcount + len(rowset)
            if len(rowset) < arraySize:
                # 已经没有什么可以取的了, 游标结束
                fetchStatus = False
        else:
            if cursor.warnings is not None:
                status = "{0} row{1} affected with warnings."
            else:
                status = "{0} row{1} affected."
            rowcount = 0 if cursor.rowcount == -1 else cursor.rowcount
            result = None
            fetchStatus = False

        # 只要不是最后一次打印，不再返回status内容
        if fetchStatus:
            status = None

        if self.testOptions.get('FEEDBACK').upper() == 'ON' and status is not None:
            status = status.format(rowcount, "" if rowcount in [0, 1] else "s")
        else:
            status = None
        return title, result, headers, columnTypes, status, fetchStatus, rowcount, cursor.warnings

    '''
        执行指定的API请求

        输入：
            apiRequest             JSON对象，请求的内容
            apiHints               提示信息            
    '''
    def executeAPIStatement(self, apiRequest, apiHints, startTime):
        """
        返回内容：
            错误情况下：
            {
                "type": "error",
                "message": apiErrorMessage
            }
            正确情况下：
            {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": content
            }
        """
        httpHandler = self.cliHandler.httpHandler

        httpMethod = apiRequest["httpMethod"]
        httpRequestTarget = apiRequest["httpRequestTarget"]

        headers = {}
        fields = {}
        if "headers" in apiRequest:
            headers = apiRequest["headers"]
        if "httpFields" in apiRequest:
            fields = apiRequest["httpFields"]

        if "contents" in apiRequest:
            body = "".join(apiRequest["contents"])
        else:
            body = ""
        try:
            # 重置Header的Content-Length
            headers["Content-Length"] = len(str(body))
            ret = httpHandler.request(
                method=httpMethod,
                url=httpRequestTarget,
                headers=headers,
                fields=fields,
                body=body
            )
            result = {"status": ret.status}
            data = ret.data.decode('utf-8')
            try:
                data = json.loads(data)
            except JSONDecodeError:
                # 返回的对象不是一个JSON
                pass
            result["content"] = data
            result = json.dumps(obj=result,
                                sort_keys=True,
                                indent=4,
                                separators=(',', ': '),
                                ensure_ascii=False)
            yield {
                "type": "result",
                "title": None,
                "rows": None,
                "headers": None,
                "columnTypes": None,
                "status": result
            }
            return
        except (
                urllib3.exceptions.MaxRetryError, urllib3.exceptions.ProtocolError) as ex:
            yield {
                "type": "error",
                "message": str(ex)
            }
            return

    def executeSQLStatement(self, sql: str, sqlHints, startTime):
        """
        返回内容：
            错误情况下：
            {
                "type": "error",
                "message": sqlErrorMessage
            }
            正确情况下：
            {
                "type": "result",
                "title": title,
                "rows": result,
                "headers": headers,
                "columnTypes": columnTypes,
                "status": status
            }
        """

        # 进入到SQL执行阶段, 开始执行SQL语句
        if self.sqlConn:
            # 打开游标
            self.sqlCursor = self.sqlConn.cursor()
        else:
            # 进入到SQL执行阶段，不是特殊命令, 数据库连接也不存在, 直接报错
            if self.testOptions.get("WHENEVER_ERROR") == "EXIT":
                raise TestCliException("Not Connected. ")
            else:
                yield {
                    "type": "error",
                    "message": "Not connected. "
                }

        # 执行正常的SQL语句
        if self.sqlCursor is not None:
            try:
                # 执行数据库的SQL语句
                if "SQL_DIRECT" in sqlHints.keys():
                    self.sqlCursor.execute_direct(sql, TimeOutLimit=self.timeout)
                elif "SQL_PREPARE" in sqlHints.keys():
                    self.sqlCursor.execute(sql, TimeOutLimit=self.timeout)
                else:
                    if self.testOptions.get("SQL_EXECUTE").upper() == "DIRECT":
                        self.sqlCursor.execute_direct(sql, TimeOutLimit=self.timeout)
                    else:
                        self.sqlCursor.execute(sql, TimeOutLimit=self.timeout)

                rowcount = 0
                sqlStatus = 0
                while True:
                    (title, result, headers, columnTypes, status,
                     fetchStatus, fetchedRowCount, sqlWarnings) = \
                        self.getcommandResult(self.sqlCursor, rowcount)
                    rowcount = fetchedRowCount
                    if "TESTCLI_DEBUG" in os.environ:
                        print("[DEBUG] headers=" + str(headers))
                        if result is not None:
                            for rowPos in range(0, len(result)):
                                for cellPos in range(0, len(result[rowPos])):
                                    print("[DEBUG] Cell[" + str(rowPos) + ":" +
                                          str(cellPos) + "]=[" + str(result[rowPos][cellPos]) + "]")

                    sqlErrorMessage = ""
                    # 如果Hints中有order字样，对结果进行排序后再输出
                    if "Order" in sqlHints.keys() and result is not None:
                        if "TESTCLI_DEBUG" in os.environ:
                            print("[DEBUG] Apply Sort for this result 2.")
                        # 不能用sorted函数，需要考虑None出现在列表中特定元素的问题
                        # l =  [(-32767,), (32767,), (None,), (0,)]
                        sortresult(result)

                    # 如果Hint中存在LogFilter，则结果集中过滤指定的输出信息
                    if "LogFilter" in sqlHints.keys() and result is not None:
                        for sqlFilter in sqlHints["LogFilter"]:
                            for item in result[:]:
                                if "TESTCLI_DEBUG" in os.environ:
                                    print("[DEBUG] Apply Filter: [" + str(''.join(str(item))) + "] with " + sqlFilter)
                                if re.match(pattern=sqlFilter, string=''.join(str(item)), flags=re.IGNORECASE):
                                    result.remove(item)
                                    continue

                    # 如果Hint中存在LogMask,则掩码指定的输出信息
                    if "LogMask" in sqlHints.keys() and result is not None:
                        for i in range(0, len(result)):
                            rowResult = list(result[i])
                            bDataChanged = False
                            for j in range(0, len(rowResult)):
                                if rowResult[j] is None:
                                    continue
                                output = str(rowResult[j])
                                for sqlMaskListString in sqlHints["LogMask"]:
                                    sqlMaskList = sqlMaskListString.split("=>")
                                    if len(sqlMaskList) == 2:
                                        sqlMaskPattern = sqlMaskList[0]
                                        sqlMaskTarget = sqlMaskList[1]
                                        if "TESTCLI_DEBUG" in os.environ:
                                            print("[DEBUG] Apply Mask: " + output +
                                                  " with " + sqlMaskPattern + "=>" + sqlMaskTarget)
                                        try:
                                            beforeReplace = str(rowResult[j])
                                            nIterCount = 0
                                            while True:
                                                # 循环多次替代，一直到没有可替代为止
                                                afterReplace = re.sub(sqlMaskPattern, sqlMaskTarget,
                                                                      beforeReplace, re.IGNORECASE)
                                                if afterReplace == beforeReplace or nIterCount > 99:
                                                    newOutput = afterReplace
                                                    break
                                                beforeReplace = afterReplace
                                                nIterCount = nIterCount + 1
                                            if newOutput != output:
                                                bDataChanged = True
                                                rowResult[j] = newOutput
                                        except re.error:
                                            if "TESTCLI_DEBUG" in os.environ:
                                                print('[DEBUG] traceback.print_exc():\n%s'
                                                      % traceback.print_exc())
                                                print('[DEBUG] traceback.format_exc():\n%s'
                                                      % traceback.format_exc())
                                    else:
                                        if "TESTCLI_DEBUG" in os.environ:
                                            print("[DEBUG] LogMask Hint Error: " + sqlHints["LogMask"])
                            if bDataChanged:
                                result[i] = tuple(rowResult)

                    # 返回SQL结果
                    if sqlStatus == 1:
                        yield {
                            "type": "error",
                            "message": sqlErrorMessage
                        }
                    else:
                        if self.testOptions.get('TERMOUT').upper() != 'OFF':
                            yield {
                                "type": "result",
                                "title": title,
                                "rows": result,
                                "headers": headers,
                                "columnTypes": columnTypes,
                                "status": status
                            }
                        else:
                            yield {
                                "type": "result",
                                "title": title,
                                "rows": [],
                                "headers": headers,
                                "columnTypes": columnTypes,
                                "status": status
                            }
                    if not fetchStatus:
                        break
            except SQLCliJDBCTimeOutException:
                # 处理超时时间问题
                if sql.upper() not in ["_EXIT", "_QUIT"]:
                    if self.timeOutMode == "SCRIPT":
                        sqlErrorMessage = "TestCli-000: Script Timeout " \
                                             "(" + str(self.scriptTimeOut) + \
                                             ") expired. Abort this command."
                    else:
                        sqlErrorMessage = "TestCli-000: SQL Timeout " \
                                             "(" + str(self.sqlTimeOut) + \
                                             ") expired. Abort this command."
                    yield {"type": "error", "message": sqlErrorMessage}
            except (SQLCliJDBCException, Exception) as je:
                sqlErrorMessage = str(je).strip()
                for errorPrefix in ["java.util.concurrent.ExecutionException:", ]:
                    if sqlErrorMessage.startswith(errorPrefix):
                        sqlErrorMessage = sqlErrorMessage[len(errorPrefix):].strip()
                for errorPrefix in ['java.sql.SQLSyntaxErrorException:',
                                    "java.sql.SQLException:",
                                    "java.sql.SQLInvalidAuthorizationSpecException:",
                                    "java.sql.SQLDataException:",
                                    "java.sql.SQLTransactionRollbackException:",
                                    "java.sql.SQLTransientConnectionException:",
                                    "java.sql.SQLFeatureNotSupportedException",
                                    "com.microsoft.sqlserver.jdbc.",
                                    "org.h2.jdbc.JdbcSQLSyntaxErrorException:",
                                    "dm.jdbc.driver.DMException:",
                                    ]:
                    if sqlErrorMessage.startswith(errorPrefix):
                        sqlErrorMessage = sqlErrorMessage[len(errorPrefix):].strip()

                # 如果Hint中存在LogFilter，则输出的消息中过滤指定的输出信息
                if "LogFilter" in sqlHints.keys():
                    sqlMultiLineErrorMessage = sqlErrorMessage.split('\n')
                    bErrorMessageHasChanged = False
                    for sqlFilter in sqlHints["LogFilter"]:
                        for item in sqlMultiLineErrorMessage[:]:
                            if re.match(sqlFilter, ''.join(str(item)), re.IGNORECASE):
                                sqlMultiLineErrorMessage.remove(item)
                                bErrorMessageHasChanged = True
                                continue
                    if bErrorMessageHasChanged:
                        sqlErrorMessage = "\n".join(sqlMultiLineErrorMessage)

                # 如果Hint中存在LogMask,则掩码指定的输出信息
                if "LogMask" in sqlHints.keys():
                    sqlMultiLineErrorMessage = sqlErrorMessage.split('\n')
                    bErrorMessageHasChanged = False
                    for sqlMaskListString in sqlHints["LogMask"]:
                        sqlMaskList = sqlMaskListString.split("=>")
                        if len(sqlMaskList) == 2:
                            sqlMaskPattern = sqlMaskList[0]
                            sqlMaskTarget = sqlMaskList[1]
                            for pos2 in range(0, len(sqlMultiLineErrorMessage)):
                                newOutput = re.sub(sqlMaskPattern, sqlMaskTarget,
                                                   sqlMultiLineErrorMessage[pos2],
                                                   re.IGNORECASE)
                                if newOutput != sqlMultiLineErrorMessage[pos2]:
                                    bErrorMessageHasChanged = True
                                    sqlMultiLineErrorMessage[pos2] = newOutput
                        else:
                            if "TESTCLI_DEBUG" in os.environ:
                                raise TestCliException("LogMask Hint Error: " + sqlHints["LogMask"])
                    if bErrorMessageHasChanged:
                        sqlErrorMessage = "\n".join(sqlMultiLineErrorMessage)
                yield {"type": "error", "message": sqlErrorMessage}

    def runStatement(self, statement: str,
                     commandScriptFile: str = "Console",
                     nameSpace: str = None):

        # 如果没有提供nameSpace，则使用系统默认的nameSpace
        if nameSpace is None:
            nameSpace = self.testOptions.get("NAMESPACE")

        # Remove spaces and EOL
        statement = statement.strip()
        formattedCommand = None
        if not statement:  # Empty string
            return

        # 记录脚本的文件名
        self.script = commandScriptFile

        # DEBUG模式下，打印当前计划要执行的语句
        if "TESTCLI_DEBUG" in os.environ:
            if nameSpace == "SQL":
                print("[DEBUG] SQL Command=[" + str(statement) + "]")
            elif nameSpace == "API":
                print("[DEBUG] API Command=[" + str(statement) + "]")

        # 开始解析语句
        try:
            ret_CommandSplitResults = []
            ret_CommandSplitResultsWithComments = []
            ret_CommandHints = []

            # 解析前要保留用户的nameSpace，解析后要把这个内容还原
            # 主要是考虑了执行脚本中，脚本内部含有的nameSpace切换不应该影响到外部的脚本
            defaultNameSpace = self.testOptions.get("NAMESPACE")

            # 将所有的语句分拆开，按照行，依次投喂给解析器，以获得被Antlr分拆后的运行结果
            currentStatement = None
            currentStatementWithComments = None
            currentHints = []
            statementLines = statement.split('\n')
            for nPos in range(0, len(statementLines)):
                statementLine = statementLines[nPos]
                # 将上次没有结束的行和当前行放在一起, 再次看是否已经结束
                if currentStatement is None:
                    currentStatement = statementLine
                else:
                    currentStatement = currentStatement + '\n' + statementLine
                if currentStatementWithComments is None:
                    currentStatementWithComments = statementLine
                else:
                    currentStatementWithComments = currentStatementWithComments + '\n' + statementLine

                # 调用解析器解析语句
                if self.testOptions.get("NAMESPACE") == "SQL":
                    (isFinished, ret_CommandSplitResult, ret_errorCode, ret_errorMsg) \
                        = SQLAnalyze(currentStatement)
                elif self.testOptions.get("NAMESPACE") == "API":
                    (isFinished, ret_CommandSplitResult, ret_errorCode, ret_errorMsg) \
                        = APIAnalyze(currentStatement)
                else:
                    raise TestCliException("不支持的运行空间【" + str(nameSpace) + "】")

                # 如果发生了语句解析错误，且错误信息是缺少EOF，则是认为语句没有结束导致，不是正常的错误
                if ret_errorCode != 0:
                    if re.search(pattern=r'missing.*<EOF>', string=ret_errorMsg):
                        # 语句没有结束
                        isFinished = False
                    if re.search(pattern=r'expecting.*<EOF>', string=ret_errorMsg):
                        # 语句没有结束
                        isFinished = False
                    if not isFinished and self.testOptions.get("NAMESPACE") == "SQL":
                        if currentStatement.strip().endswith(';') or currentStatement.strip().endswith("\n/"):
                            if currentStatement.strip().startswith("_"):
                                # 内部语句，且语句已经结束
                                ret_CommandSplitResults.append(
                                    {'name': 'PARSE_ERROR',
                                     'statement': currentStatement,
                                     'reason': ret_errorMsg}
                                )
                                # 解析前的语句
                                ret_CommandSplitResultsWithComments.append(currentStatementWithComments)
                                # 所有的提示信息
                                ret_CommandHints.append(currentHints)
                                # 清空语句的变量
                                currentHints = []
                                currentStatement = None
                                currentStatementWithComments = None
                            else:
                                ret_CommandSplitResults.append(
                                    {'name': 'SQL_UNKNOWN',
                                     'statement': currentStatement,
                                     'reason': ret_errorMsg}
                                )
                                # 解析前的语句
                                ret_CommandSplitResultsWithComments.append(currentStatementWithComments)
                                # 所有的提示信息
                                ret_CommandHints.append(currentHints)
                                # 清空语句的变量
                                currentHints = []
                                currentStatement = None
                                currentStatementWithComments = None
                        continue

                # 如果语句没有结束，则需要等待下一句输入
                if not isFinished:
                    # 如果到了文件末尾，就没有必要继续等待，直接返回
                    if nPos == (len(statementLines) - 1):
                        if self.testOptions.get("NAMESPACE") == "SQL":
                            ret_CommandSplitResults.append(
                                {'name': 'SQL_UNKNOWN',
                                 'statement': currentStatement,
                                 'reason': ret_errorMsg}
                            )
                        elif self.testOptions.get("NAMESPACE") == "API":
                            ret_CommandSplitResults.append(
                                {'name': 'API_UNKNOWN',
                                 'statement': currentStatement,
                                 'reason': ret_errorMsg}
                            )
                        # 解析前的语句
                        ret_CommandSplitResultsWithComments.append(currentStatementWithComments)
                        # 所有的提示信息
                        ret_CommandHints.append(currentHints)
                        # 清空语句的变量
                        currentHints = []
                        currentStatement = None
                        currentStatementWithComments = None
                    continue

                # 语句已经结束, 但是输入的内容有错误信息
                if isFinished and ret_errorCode != 0:
                    ret_CommandSplitResults.append(
                        {'name': 'PARSE_ERROR',
                         'statement': currentStatement,
                         'reason': ret_errorMsg}
                    )
                    # 解析前的语句
                    ret_CommandSplitResultsWithComments.append(currentStatementWithComments)
                    # 所有的提示信息
                    ret_CommandHints.append(currentHints)
                    # 清空语句的变量
                    currentHints = []
                    currentStatement = None
                    currentStatementWithComments = None
                    continue

                # 语句已经结束，没有任何错误，但解析的结果为空
                if ret_CommandSplitResult is None:
                    # 空行，可能包含注释，保留注释内容，带到下一个有意义的段落
                    # Hint的两种写法
                    # 1.  -- [Hint]  hintsomething
                    # 2.  -- [hitsomething]
                    pattern = r"^(\s+)?([-/][-/])(\s+)?\[(\s+)?Hint(\s+)?\](.*)"
                    matchObj = re.match(pattern, statementLine, re.IGNORECASE)
                    if matchObj:
                        currentHints.append(matchObj.group(6).strip())
                    else:
                        pattern = r"^(\s+)?([-/][-/])(\s+)?\[(.*)\].*"
                        matchObj = re.match(pattern, statementLine, re.IGNORECASE)
                        if matchObj:
                            currentHints.append(matchObj.group(4).strip())
                    # 解析后的语句
                    ret_CommandSplitResults.append(None)
                    # 解析前的语句
                    ret_CommandSplitResultsWithComments.append(currentStatementWithComments)
                    # 对于非有效语句，Hint不在当前语句中体现，而是要等到下次有意义的语句进行处理
                    ret_CommandHints.append([])
                    # 清空除了Hint以外的其他语句变量
                    currentStatement = None
                    currentStatementWithComments = None
                    continue

                # 切换解析的命名空间
                if ret_CommandSplitResult["name"] == "USE":
                    self.testOptions.set("NAMESPACE", ret_CommandSplitResult["nameSpace"])

                # 对于SQL的CONNECT命令做特殊处理， 以支持外部环境变量带来的问题
                if self.testOptions.get("NAMESPACE") == "SQL":
                    if ret_CommandSplitResult["name"] == "CONNECT":
                        # 对于数据库连接命令，如果没有给出连接详细信息，并且指定了环境变量，附属环境变量到连接命令后
                        if "driver" not in ret_CommandSplitResult:
                            connectionURL = None
                            if "SQLCLI_CONNECTION_URL" in os.environ:
                                connectionURL = str(os.environ["SQLCLI_CONNECTION_URL"]).strip('"').strip("'").strip()
                            elif "TESTCLI_CONNECTION_URL" in os.environ:
                                connectionURL = str(os.environ["TESTCLI_CONNECTION_URL"]).strip('"').strip("'").strip()
                            if connectionURL is not None:
                                # 拼接链接字符串后重新解析
                                (isFinished, ret_CommandSplitResult, ret_errorCode, ret_errorMsg) \
                                    = SQLAnalyze(currentStatement + "@" + connectionURL)

                # 语句正常结束，且其中包含了正式的内容
                ret_CommandSplitResults.append(ret_CommandSplitResult)
                # 解析前的语句
                ret_CommandSplitResultsWithComments.append(currentStatementWithComments)
                # 所有的提示信息
                ret_CommandHints.append(currentHints)
                # 清空语句的变量
                currentHints = []
                currentStatement = None
                currentStatementWithComments = None
                continue
        except Exception:
            if "TESTCLI_DEBUG" in os.environ:
                print('traceback.print_exc():\n%s' % traceback.print_exc())
                print('traceback.format_exc():\n%s' % traceback.format_exc())
            raise TestCliException("TestCli-000 Internal error. Parse failed.")

        # 解析过程中设置的NAMESPACE，解析后要还原，好保证执行的正确
        self.testOptions.set("NAMESPACE", defaultNameSpace)

        # 开始执行语句
        pos = 0
        while True:
            if pos == len(ret_CommandSplitResults):
                break

            # 记录命令开始时间
            startTime = time.time()
            try:
                self.sqlTimeOut = int(self.testOptions.get("SQL_TIMEOUT"))
            except ValueError:
                self.sqlTimeOut = -1
            try:
                self.apiTimeOut = int(self.testOptions.get("API_TIMEOUT"))
            except ValueError:
                self.apiTimeOut = -1
            try:
                self.scriptTimeOut = int(self.testOptions.get("SCRIPT_TIMEOUT"))
            except ValueError:
                self.scriptTimeOut = -1

            # 首先打印原有语句
            if self.testOptions.get("NAMESPACE") == "SQL":
                formattedCommand = SQLFormatWithPrefix(ret_CommandSplitResultsWithComments[pos])
            if self.testOptions.get("NAMESPACE") == "API":
                formattedCommand = APIRequestStringFormatWithPrefix(ret_CommandSplitResultsWithComments[pos])

            # 如果是空语句，不需要执行，但可能是完全注释行
            # 也可能是一个解析错误的语句
            if ret_CommandSplitResults[pos] is None:
                if self.loopMode and not self.loopCondition:
                    pass
                else:
                    yield {
                        "type": "parse",
                        "rawCommand": ret_CommandSplitResults[pos],
                        "formattedCommand": formattedCommand,
                        "rewrotedCommand": [],
                        "script": commandScriptFile
                    }
                pos = pos + 1
                continue
            else:
                parseObject = dict(ret_CommandSplitResults[pos])

            # 如果在循环模式下，且循环条件已经不满足，则直接跳出
            if parseObject["name"] != "LOOP":
                if self.loopMode and not self.loopCondition:
                    pos = pos + 1
                    continue

            # 返回Command的解析信息
            yield {
                "type": "parse",
                "rawCommand": ret_CommandSplitResults[pos],
                "formattedCommand": formattedCommand,
                "rewrotedCommand": [],
                "script": commandScriptFile
            }

            # 处理超时时间问题
            if self.scriptTimeOut > 0:
                if self.scriptTimeOut <= time.time() - self.getStartTime():
                    commandErrorMessage = "TestCli-000: Script Timeout " \
                                         "(" + str(round(self.scriptTimeOut, 2)) + \
                                         ") expired. Abort this Script."
                    yield {"type": "error", "message": commandErrorMessage}
                    raise EOFError
                else:
                    if self.testOptions.get("NAMESPACE") == "SQL":
                        if self.sqlTimeOut > 0:
                            if self.scriptTimeOut - (time.time() - self.getStartTime()) < self.sqlTimeOut:
                                # 脚本超时剩余时间更少，执行较少的那个超时控制
                                self.timeOutMode = "SCRIPT"
                                self.timeout = self.scriptTimeOut - (time.time() - self.getStartTime())
                            else:
                                self.timeOutMode = "COMMAND"
                                self.timeout = self.sqlTimeOut
                        else:
                            self.timeOutMode = "SCRIPT"
                            self.timeout = self.scriptTimeOut - (time.time() - self.getStartTime())
                    if self.testOptions.get("NAMESPACE") == "API":
                        if self.apiTimeOut > 0:
                            if self.scriptTimeOut - (time.time() - self.getStartTime()) < self.apiTimeOut:
                                # 脚本超时剩余时间更少，执行较少的那个超时控制
                                self.timeOutMode = "SCRIPT"
                                self.timeout = self.scriptTimeOut - (time.time() - self.getStartTime())
                            else:
                                self.timeOutMode = "COMMAND"
                                self.timeout = self.apiTimeOut
                        else:
                            self.timeOutMode = "SCRIPT"
                            self.timeout = self.scriptTimeOut - (time.time() - self.getStartTime())
            elif self.testOptions.get("NAMESPACE") == "SQL" and self.sqlTimeOut > 0:
                # 没有设置SCRIPT的超时时间，只设置了COMMAND的超时时间
                self.timeOutMode = "COMMAND"
                self.timeout = self.sqlTimeOut
            elif self.testOptions.get("NAMESPACE") == "API" and self.apiTimeOut > 0:
                # 没有设置SCRIPT的超时时间，只设置了COMMAND的超时时间
                self.timeOutMode = "COMMAND"
                self.timeout = self.apiTimeOut
            else:
                # 什么超时时间都没有设置
                self.timeOutMode = None
                self.timeout = -1

            # 处理Hints信息
            if self.testOptions.get("NAMESPACE") == "SQL":
                commandHintList = parseSQLHints(list(ret_CommandHints[pos]))
            elif self.testOptions.get("NAMESPACE") == "API":
                commandHintList = parseAPIHints(list(ret_CommandHints[pos]))
            else:
                commandHintList = {}
            for commandHintKey, commandHintValue in commandHintList.items():
                # 根据语句中的变量或者其他定义信息来重写当前语句
                if type(commandHintValue) == str:
                    # 只针对字符串进行处理，暂时不对其他数据类型进行处理
                    commandHintNewValue, rewrotedCommandHintValueList = rewriteHintStatement(
                        cls=self.cliHandler,
                        statement=commandHintValue,
                        commandScriptFile=commandScriptFile
                    )
                    if len(rewrotedCommandHintValueList) != 0:
                        commandHintList[commandHintKey] = commandHintNewValue
                        yield {
                            "type": "parse",
                            "rawCommand": None,
                            "formattedCommand": None,
                            "rewrotedCommand":
                                ["REWROTED Hint> " + str(commandHintKey) + ":" + str(commandHintNewValue), ],
                            "script": commandScriptFile
                        }
            if "ScenarioName" in commandHintList:
                scenarioName = str(commandHintList["ScenarioName"])
                scenarioId = str(commandHintList["ScenarioId"])
                if scenarioName.strip().upper() == 'END':
                    self.scenarioName = ""
                    self.scenarioId = ""
                else:
                    self.scenarioName = scenarioName
                    self.scenarioId = scenarioId

            # 处理各种命令
            if "TESTCLI_DEBUG" in os.environ:
                print("[DEBUG] parsedObject=[" + str(parseObject) + "]")
            sqlKeyWords = ["SELECT", "DELETE", "UPDATE", "CREATE", "INSERT",
                           "DROP", "COMMIT", "ROLLBACK",
                           "PROCEDURE", "DECLARE", "BEGIN",
                           "SQL_UNKNOWN"]
            if parseObject["name"] == "HELP":
                # 显示帮助信息
                for commandResult in showHelp(
                        cls=self.cliHandler,
                        topicName=parseObject["topic"],
                ):
                    yield commandResult
            elif parseObject["name"] == "ECHO":
                # 将后续内容回显到指定的文件中
                for commandResult in echo_input(
                        cls=self.cliHandler,
                        fileName=parseObject["param"],
                        block=parseObject["block"],
                ):
                    yield commandResult
            elif parseObject["name"] == "START":
                # 执行脚本前记录当前执行的脚本名称
                savedExecuteScript = self.cliHandler.executeScript
                self.cliHandler.executeScript = parseObject["script"]
                # 执行脚本文件
                for commandResult in executeFile(
                        cls=self.cliHandler,
                        scriptFile=parseObject["script"],
                        argv=parseObject["argv"],
                ):
                    yield commandResult
                # 执行脚本后还原当前执行脚本的名称
                self.cliHandler.executeScript = savedExecuteScript
            elif parseObject["name"] in ["EXIT", "QUIT"]:
                # 执行脚本文件
                if "exitValue" in parseObject.keys():
                    exitValue = parseObject["exitValue"]
                else:
                    exitValue = 0
                for commandResult in exitApplication(
                        cls=self.cliHandler,
                        exitValue=exitValue
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] == "CONNECT":
                # 执行CONNECT命令
                if self.testOptions.get("NAMESPACE") == "SQL":
                    for commandResult in connectDb(
                            cls=self.cliHandler,
                            connectProperties=parseObject
                    ):
                        if commandResult["type"] == "result":
                            lastCommandResult.clear()
                            lastCommandResult["status"] = commandResult["status"]
                            lastCommandResult["errorCode"] = 0
                        if commandResult["type"] == "error":
                            lastCommandResult["message"] = commandResult["message"]
                            lastCommandResult["errorCode"] = 1
                        yield commandResult
                else:
                    lastCommandResult["message"] = "Not support connect command in non-sql namespace"
                    lastCommandResult["errorCode"] = 1
                    yield {
                        "type": "error",
                        "message": "Not support connect command in non-sql namespace",
                        "script": commandScriptFile
                    }
            elif parseObject["name"] == "SET":
                # 执行SET命令
                for commandResult in setOptions(
                        cls=self.cliHandler,
                        options=parseObject
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] == "DISCONNECT":
                # 执行DISCONNECT命令
                if self.testOptions.get("NAMESPACE") == "SQL":
                    for commandResult in disconnectDb(
                            cls=self.cliHandler
                    ):
                        if commandResult["type"] == "result":
                            lastCommandResult.clear()
                            lastCommandResult["status"] = commandResult["status"]
                            lastCommandResult["errorCode"] = 0
                        if commandResult["type"] == "error":
                            lastCommandResult["status"] = commandResult["message"]
                            lastCommandResult["errorCode"] = 1
                        yield commandResult
                else:
                    yield {
                        "type": "error",
                        "message": "affadsfsad",
                        "script": commandScriptFile
                    }
            elif parseObject["name"] in sqlKeyWords:
                if self.ifMode and not self.ifCondition:
                    pos = pos + 1
                    continue
                sqlCommand = parseObject["statement"]
                # 根据语句中的变量或者其他定义信息来重写当前语句
                sqlCommand, rewrotedCommandList = rewriteSQLStatement(
                    cls=self.cliHandler,
                    statement=sqlCommand,
                    commandScriptFile=commandScriptFile
                )
                if len(rewrotedCommandList) != 0:
                    # 如果命令被发生了改写，要打印改写记录
                    yield {
                        "type": "parse",
                        "rawCommand": None,
                        "formattedCommand": None,
                        "rewrotedCommand": rewrotedCommandList,
                        "script": commandScriptFile
                    }

                # 执行SQL语句
                for result in self.executeSQLStatement(
                        sql=sqlCommand,
                        sqlHints=commandHintList,
                        startTime=0):
                    if result["type"] == "result":
                        lastCommandResult["rows"] = result["rows"]
                        lastCommandResult["headers"] = result["headers"]
                        lastCommandResult["status"] = result["status"]
                        lastCommandResult["errorCode"] = 0
                    if result["type"] == "error":
                        lastCommandResult["rows"] = []
                        lastCommandResult["headers"] = []
                        lastCommandResult["status"] = result["message"]
                        lastCommandResult["errorCode"] = 1
                    yield result
            elif parseObject["name"] in ["USE"]:
                for commandResult in userNameSpace(
                        cls=self.cliHandler,
                        nameSpace=parseObject["nameSpace"]
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] in ["SLEEP"]:
                for commandResult in cliSleep(
                        cls=self.cliHandler,
                        sleepTime=parseObject["sleepTime"]
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] in ["SPOOL"]:
                for commandResult in spool(
                        cls=self.cliHandler,
                        fileName=parseObject["file"]
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] in ["SESSION"]:
                for commandResult in sessionManage(
                        cls=self.cliHandler,
                        action=parseObject["action"],
                        sessionName=parseObject["sessionName"]
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] in ["SCRIPT"]:
                for commandResult in executeEmbeddScript(
                        cls=self.cliHandler,
                        block=parseObject["block"]
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] in ["ASSERT"]:
                for commandResult in assertExpression(
                        cls=self.cliHandler,
                        expression=parseObject["expression"],
                        assertName=parseObject["assertName"]
                ):
                    if commandResult["type"] == "result":
                        lastCommandResult.clear()
                        lastCommandResult["status"] = commandResult["status"]
                        lastCommandResult["errorCode"] = 0
                    if commandResult["type"] == "error":
                        lastCommandResult["status"] = commandResult["message"]
                        lastCommandResult["errorCode"] = 1
                    yield commandResult
            elif parseObject["name"] in ["LOAD"]:
                if parseObject["option"] == "PLUGIN":
                    for commandResult in loadPlugin(
                            cls=self.cliHandler,
                            pluginFile=parseObject["pluginFile"]
                    ):
                        if commandResult["type"] == "result":
                            lastCommandResult.clear()
                            lastCommandResult["status"] = commandResult["status"]
                            lastCommandResult["errorCode"] = 0
                        if commandResult["type"] == "error":
                            lastCommandResult["status"] = commandResult["message"]
                            lastCommandResult["errorCode"] = 1
                        yield commandResult
                elif parseObject["option"] == "JDBCDRIVER":
                    driverName = None
                    driverClass = None
                    driverFile = None
                    driverURL = None
                    driverProps = None
                    if "driverName" in parseObject:
                        driverName = parseObject["driverName"]
                    if "driverClass" in parseObject:
                        driverClass = parseObject["driverClass"]
                    if "driverFile" in parseObject:
                        driverFile = parseObject["driverFile"]
                    if "driverURL" in parseObject:
                        driverURL = parseObject["driverURL"]
                    if "driverProps" in parseObject:
                        driverProps = parseObject["driverProps"]
                    for commandResult in loadJDBCDriver(
                            cls=self.cliHandler,
                            driverName=driverName,
                            driverClass=driverClass,
                            driverFile=driverFile,
                            driverURL=driverURL,
                            driverProps=driverProps
                    ):
                        if commandResult["type"] == "result":
                            lastCommandResult.clear()
                            lastCommandResult["status"] = commandResult["status"]
                            lastCommandResult["errorCode"] = 0
                        if commandResult["type"] == "error":
                            lastCommandResult["status"] = commandResult["message"]
                            lastCommandResult["errorCode"] = 1
                        yield commandResult
                elif parseObject["option"] == "MAP":
                    for commandResult in loadMap(
                            cls=self.cliHandler,
                            mapFile=parseObject["mapFile"]
                    ):
                        if commandResult["type"] == "result":
                            lastCommandResult.clear()
                            lastCommandResult["status"] = commandResult["status"]
                            lastCommandResult["errorCode"] = 0
                        if commandResult["type"] == "error":
                            lastCommandResult["status"] = commandResult["message"]
                            lastCommandResult["errorCode"] = 1
                        yield commandResult
            elif parseObject["name"] in ["HTTP"]:
                if self.ifMode and not self.ifCondition:
                    pos = pos + 1
                    continue
                # 根据语句中的变量或者其他定义信息来重写当前语句
                httpRequestTarget, rewrotedCommandList = rewriteAPIStatement(
                    cls=self.cliHandler,
                    requestObject=parseObject,
                    commandScriptFile=commandScriptFile
                )
                if len(rewrotedCommandList) != 0:
                    # 如果命令被发生了改写，要打印改写记录
                    yield {
                        "type": "parse",
                        "rawCommand": None,
                        "formattedCommand": None,
                        "rewrotedCommand": rewrotedCommandList,
                        "script": commandScriptFile
                    }

                # 执行HTTP请求
                for result in self.executeAPIStatement(
                        apiRequest=parseObject,
                        apiHints=commandHintList,
                        startTime=0):
                    if result["type"] == "result":
                        lastCommandResult.clear()
                        data = json.loads(result["status"])
                        lastCommandResult["content"] = data["content"]
                        lastCommandResult["status"] = data["status"]
                        lastCommandResult["errorCode"] = 0
                    if result["type"] == "error":
                        lastCommandResult["message"] = result["message"]
                        lastCommandResult["errorCode"] = 1
                    yield result
            elif parseObject["name"] in ["HOST"]:
                # 执行主机操作系统命令
                for result in executeLocalCommand(
                        cls=self.cliHandler,
                        command=parseObject["script"]
                ):
                    yield result
            elif parseObject["name"] in ["ENDIF"]:
                self.ifMode = False
                self.ifCondition = False
            elif parseObject["name"] in ["IF"]:
                expression = parseObject["expression"]
                try:
                    ret = evalExpression(self.cliHandler, expression)
                    if type(ret) == bool:
                        self.ifMode = True
                        if ret:
                            self.ifCondition = True
                        else:
                            self.ifCondition = False
                    else:
                        yield {
                            "type": "error",
                            "message": "Set condition fail. SyntaxError =>[not a bool expression]"
                        }
                except (SyntaxError, NameError) as ae:
                    yield {
                        "type": "error",
                        "message": "Set condition fail. SyntaxError =>[" + str(ae) + "]"
                    }
            elif parseObject["name"] in ["LOOP"]:
                if parseObject["rule"] == "END":
                    if not self.loopCondition:
                        self.loopMode = False
                    else:
                        pos = self.loopStartPos
                        continue
                elif parseObject["rule"] == "BREAK":
                    self.loopCondition = False
                elif parseObject["rule"] == "CONTINUE":
                    pos = self.loopStartPos
                    continue
                elif parseObject["rule"] == "BEGIN":
                    self.loopMode = True
                    self.loopStartPos = pos
                    self.loopCondition = not evalExpression(self.cliHandler, parseObject["UNTIL"])
            elif parseObject["name"] in ["WHENEVER"]:
                for result in setWheneverAction(
                        cls=self.cliHandler,
                        action=parseObject["action"],
                        exitCode=parseObject["exitCode"]
                ):
                    yield result
            elif parseObject["name"] in ["SSH"]:
                parseObject, rewrotedCommandList = rewriteSshRequest(
                    cls=self.cliHandler,
                    requestObject=parseObject,
                    commandScriptFile=commandScriptFile
                )
                if len(rewrotedCommandList) != 0:
                    # 如果命令被发生了改写，要打印改写记录
                    yield {
                        "type": "parse",
                        "rawCommand": None,
                        "formattedCommand": None,
                        "rewrotedCommand": rewrotedCommandList,
                        "script": commandScriptFile
                    }

                for result in executeSshRequest(
                        requestObject=parseObject,
                ):
                    yield result
            elif parseObject["name"] in ["JOB"]:
                for result in self.cliHandler.JobHandler.processRequest(
                        cls=self.cliHandler,
                        requestObject=parseObject,
                ):
                    yield result
            elif parseObject["name"] in ["COMPARE"]:
                for result in executeCompareRequest(
                        requestObject=parseObject,
                ):
                    yield result
            elif parseObject["name"] in ["DATA"]:
                for result in executeDataRequest(
                        cls=self.cliHandler,
                        requestObject=parseObject,
                ):
                    yield result
            elif parseObject["name"] in ["PARSE_ERROR"]:
                yield {"type": "error",
                       "message": "TestCli parse error:  " + str(parseObject["reason"])}
            elif parseObject["name"] in ["MONITOR"]:
                for result in executeMonitorRequest(
                        cls=self.cliHandler,
                        requestObject=parseObject,
                ):
                    yield result
            else:
                raise TestCliException("FDDFSFDFSDSFD " + str(parseObject))

            # 如果需要，打印语句执行时间
            endTime = time.time()
            lastCommandResult["elapsed"] = endTime - startTime

            # 记录命令的日志信息
            if self.testOptions.get("SILENT").upper() == 'OFF':
                if parseObject["name"] == "START":
                    commandStatus = ""
                    commandErrorCode = ""
                else:
                    if "status" in lastCommandResult:
                        commandStatus = lastCommandResult["status"]
                    else:
                        commandStatus = ""
                    if "errorCode" in lastCommandResult:
                        commandErrorCode = lastCommandResult["errorCode"]
                    else:
                        commandErrorCode = ""
                yield {
                    "type": "statistics",
                    "startedTime": startTime,
                    "elapsed": endTime - startTime,
                    "processName": self.workerName,
                    "rawCommand": ret_CommandSplitResultsWithComments[pos],
                    "commandType": parseObject["name"],
                    "command": json.dumps(obj=parseObject, sort_keys=True, ensure_ascii=False),
                    "commandStatus": commandStatus,
                    "errorCode": commandErrorCode,
                    "scenarioId": self.scenarioId,
                    "scenarioName": self.scenarioName
                }

            # 开始执行下一个语句
            pos = pos + 1
